
--Camera		= require "scripts/camera2"
utf8		= require("utf8")
tools		= require "scripts/tools"
cam			= require "scripts/camera"
gameloop	= require "scripts/gameloop"
debug		= require "scripts/debug"
objmgr		= require "scripts/objmgr"
player		= require "scripts/entities/player"
zombie		= require "scripts/entities/zombie"
bat			= require "scripts/entities/bat"
spider		= require "scripts/entities/spider"
bullet		= require "scripts/entities/bullet"
--block		= require "scripts/block"		--> map
item		= require "scripts/item"
renderer	= require "scripts/renderer"
button		= require "scripts/button"
entry		= require "scripts/entry"
label		= require "scripts/label"
regulator	= require "scripts/regulator"
guimgr		= require "scripts/guimgr"
mouse		= require "scripts/mouse"
--ui			= require "scripts/userinterface"
main_menu	= require "scripts/main_menu"
lobby		= require "scripts/lobby"
settings	= require "scripts/settings"
game		= require "scripts/game"
app			= require "scripts/app"
imageloader	= require "scripts/imageloader"
soundmgr	= require "scripts/soundmgr"
network		= require "scripts/network"
userinput	= require "scripts/userinput"


-- init some globals, TODO: move somewhere else maybe?
GAMETIME = 0
AVG_DT = 0
-- mob behaviour
BEHAVIOUR_WANDER_AROUND	= 0
BEHAVIOUR_FOLLOW_PLAYER	= 1
BEHAVIOUR_DO_CIRCLES	= 2

--GAMESTATE = "ingame"

-- human player!
--localplayer = {id = 1, displayname = "Max", connection = nil, skin = nil}

-- used for stencil function to create simple terrain transitions
mask_shader = love.graphics.newShader[[
	vec4 effect (vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords) {
		if (Texel(texture, texture_coords).rgb == vec3(0.0)) {
			// a discarded pixel wont be applied as the stencil.
			discard;
		}
		return vec4(1.0);
	}
]]

test_shader = love.graphics.newShader[[
	vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords ){
		vec4 pixel = Texel(texture, texture_coords );//This is the current pixel color
		number average = (pixel.r+pixel.b+pixel.g)/3.0;
		pixel.r = average;
		pixel.g = 0;
		pixel.b = average;
		return pixel;
	}
]]

function love.load()

	math.randomseed(os.time())
	
	-- init gameloop
	gameloop:load()

	-- init app
	app:load()
	
	-- init renderer
	renderer:load()
	
	-- manage objects
	objmgr:load()
	
	-- ui controller
	--ui:load()
	
	-- manage UI objects
	guimgr:load()
	
	-- mouse
	mouse:load()
	
	-- init the user input (keyboard, mouse???)
	userinput:load()
	
	-- debugging, does not work ... no variable (string) update
	--debug:load()
	
	screen = { height = love.graphics.getHeight(), width = love.graphics.getWidth() }

	--imageloader:init()
	
	main_menu:init()
	lobby:load()
	settings:load()
	--network:init()
	game:init()
	
	-- set background color
	--love.graphics.setBackgroundColor(0.4, 0.5, 0.7)
	
	guimgr:changeUIstate("main_menu")
	
	soundmgr.playSound(soundmgr.sounds["music"]["music1"], true)
	--soundmgr.sounds["music1"]:setVolume(0.7)

	--guimgr:disableButton("btnSettingsToggleFullscreen")
	
	soundmgr.updateVolume("game")
	soundmgr.updateVolume("music")
	
end


function love.draw()
	
	--cam:set()

--	love.graphics.setShader(test_shader)

	-- background image? awesome!
	love.graphics.draw(imageloader.images["stone2"][1], 0, 0, 0, 20, 20)
	
	-- actually draw everything that has a renderer, camera set and unset moved to renderer to draw GUI
	renderer:draw()

	--cam:unset()
	
--	love.graphics.setShader()
	
	-- "mask"; looks decent with a good image, should only be drawn when ingame
	--[[
	love.graphics.setColor(1.0, 1.0, 1.0, 0.5)
	love.graphics.draw(imageloader.images["main_mask1"][1], 0, 0, 0, 1, 1)
	love.graphics.setColor(1.0, 1.0, 1.0, 1.0)
	--]]
	
	-- debug and stuff
	love.graphics.print("Current FPS: "..tostring(love.timer.getFPS( )), 10, 12)
	love.graphics.print("current dt: ".. string.sub(tostring(AVG_DT), 0, 5), 10, 24)
	--love.graphics.print("UI state: "..guimgr:getUIstate(), 10, 24)
	
	--local x, y = Mouse:getPosition()
	local wx, wy = mouse:getWorldPosition()
	--love.graphics.print("mouse pos world: " .. wx .." : " .. wy, 10, 36)
	--love.graphics.print("mouse pos local: " .. x .." : " .. y, 10, 36)
	
	--[[
	-- cam:getPosition(-mouseX, -mouseY) returns the mouse position relative to the screen center
	local cx, cy = cam:getPosition(-x, -y)
	love.graphics.print("camera pos:   " .. cx .. " : " .. cy, 10, 48)
	--
	local player = objmgr:getObjectByID("player1")
	if player ~= nil then
		local px, py = player:getPosition()
		love.graphics.print("player pos:   " .. px .. " : " .. py, 10, 60)
		
		local cx, cy = cam:getPosition()
		love.graphics.print("camera pos:   " .. cx .. " : " .. cy, 10, 48)
	end
	--]]
	
end


function love.update(dt)
	if dt > 0.1 then
		print("dt took too long: " .. dt)
		return
	end
	
	GAMETIME = GAMETIME + dt
	AVG_DT = (AVG_DT + dt) / 2	-- TODO
	gameloop:update(dt)
	soundmgr.update(dt)	-- needs .

	-- mouse pos
	--local x, y = love.mouse.getPosition()
	--print("mouse :" .. x, y)
end

--[[
function love.mousepressed( x, y, button, istouch, presses )
	
	mouse:pressed( x, y, button, istouch, presses )
	
end
--]]



