local App = {}
local APPTYPE_SOLO = 0
local APPTYPE_SERVER = 1
local APPTYPE_CLIENT = 2

function App:load()
	--print("app created")
	self.gamestate = "idle"
	self.type = APPTYPE_SOLO
	--self.isClient = false
	self.settings = {
		fullscreen = false,
		borderless = false,
		soundValues = {music = 0.2, game = 1.0},
	}
	self.humanPlayer = {id = 0, displayname = "Max", ip = 0, port = 0, skin = 1}
	self.serverAddress = "127.0.0.1"	-- TODO ... (not in use yet)
	gameloop:addLoop(self)
	
end

function App:tick(dt)
	--print("app: sound " ..self.settings.soundValueAll)
end


function App:setSolo()
	self.type = APPTYPE_SOLO
	self:updatePlayer({id = 1})
end

function App:setServer()
	self.type = APPTYPE_SERVER
end

function App:setClient()
	self.type = APPTYPE_CLIENT
end

function App:isHost()
	return self.type == APPTYPE_SERVER
end

function App:getType()
	return self.type
end


function App:soundIncrease(_category, _value)
	self.settings.soundValues[_category] = self.settings.soundValues[_category] + _value
	if self.settings.soundValues[_category] > 1 then
		self.settings.soundValues[_category] = 1
	end
	--[[
	if _category == "all" then
		self.settings.soundValueAll = self.settings.soundValueAll + _value
		if self.settings.soundValueAll > 1 then
			self.settings.soundValueAll = 1
		end
	end
	--]]
	
	-- TODO: send info to sound manager; update all sound volumes in this category, fetch value from app inside soundmanager
	soundmgr.updateVolume(_category)
end

function App:soundDecrease(_category, _value)
	self.settings.soundValues[_category] = self.settings.soundValues[_category] - _value
	if self.settings.soundValues[_category] < 0 then
		self.settings.soundValues[_category] = 0
	end

	soundmgr.updateVolume(_category)
end

-- updates information about the human player
function App:updatePlayer(_playerTable)
	self.humanPlayer.id = _playerTable.id or self.humanPlayer.id
	self.humanPlayer.displayname = _playerTable.displayname or self.humanPlayer.displayname
	--if self.humanPlayer.displayname == "" then
		--self.humanPlayer.displayname = "player"
	--end
	self.humanPlayer.ip = _playerTable.ip or self.humanPlayer.ip
	self.humanPlayer.port = _playerTable.port or self.humanPlayer.port
	self.humanPlayer.skin = _playerTable.skin or self.humanPlayer.skin
	
	--print("app: update player:")
	--print(self.humanPlayer.id)
	--print(self.humanPlayer.displayname)
	--print(self.humanPlayer.ip)
	--print(self.humanPlayer.port)
	--print("app string len: " .. string.len(self.humanPlayer.displayname))
end

function App:getHumanPlayer()
	return self.humanPlayer
end


function App:setGameState(state)
	self.gamestate = state
end

function App:getGamestate()
	return self.gamestate
end


-- called from settings
function App:toggleFullscreen()
	self.settings.fullscreen = not self.settings.fullscreen
	local width, height = 800, 600	-- love.graphics.getDimensions()
	local success = love.window.updateMode(width, height, {fullscreen = self.settings.fullscreen} )
	if success then
		screen = { height = love.graphics.getHeight(), width = love.graphics.getWidth() }	-- update these values for the camera
		print("fullscreen mode updated: " .. tostring(self.settings.fullscreen))
		print(love.graphics.getWidth())
		print(love.graphics.getHeight())
	end
end


function App:toggleBorderless()
	self.settings.borderless = not self.settings.borderless
	local width, height = love.graphics.getDimensions()
	local success = love.window.updateMode(width, height, {borderless = self.settings.borderless} )
	if success then
		screen = { height = love.graphics.getHeight(), width = love.graphics.getWidth() }	-- update these values for the camera
		print("borderless mode updated: " .. tostring(self.settings.borderless))
		print(love.graphics.getWidth())
		print(love.graphics.getHeight())
	end
end


-- called from lobby
function App:backToMainMenu()

	-- playername
	local human = app:getHumanPlayer()
	if string.len(human.displayname) == 0 then
		app:updatePlayer({displayname = "player"})
	end
	
	guimgr:changeUIstate("main_menu")
	--lobby:removePlayer(human)
	lobby:destroy()
	-- host vs guest?
	
	if network then
		network:destroy()	-- TODO: gets called from the settings ... what?
	end
	
	game:destroy()
	
end


function App:backToGame()

	guimgr:changeUIstate("ingame")
	
end


function App:getPlayername()
	return self.humanPlayer.displayname
end

function App:setPlayername()
	
	local uiElement = guimgr:getElementByID("entrySettingsPlayername")
	local text = uiElement:getText() or "?"
	self:updatePlayer( {displayname = text} )

	--[[
	local focusID = userinput:getFocus()
	if focusID and focusID == "entrySettingsPlayername" then
		local uiElement = guimgr:getElementByID(focusID)
		local text = uiElement:getText() or "?"
		app:updatePlayer( {displayname = text} )
	end
	--]]
end


function App:setServerAddress()
	local uiElement = guimgr:getElementByID("entrySettingsServerAddress")
	local text = uiElement:getText() or "?"
	--TODO: validate ip address!
	self.serverAddress = text
end


function App:getServerAddress()
	return self.serverAddress
end


-- called from main menu
function App:openLobby()

	guimgr:changeUIstate("lobby")
	lobby:create()
	self:updatePlayer( {id = 1} )	-- host has always ID 1
	self:setServer()
	--lobby:addPlayer(app:getHumanPlayer())
	
	network:startServer()
	
end


function App:joinLobby()

	guimgr:changeUIstate("lobby")
	lobby:create()
	guimgr:disableButton("btnLobbyStartGame")
	guimgr:disableButton("btnStartMP")	-- TODO: temporary "solution"; see problem with udp/socket in network.lua
	self:setClient()
	--lobby:addPlayer(app:getHumanPlayer())
	
	network:startClient()
	
end


function App:openSingleplayerGame()

	guimgr:changeUIstate("lobby")
	lobby:create()
	self:setSolo()
	lobby:addPlayer(self:getHumanPlayer())
	
end


function App:openSettings()

	guimgr:changeUIstate("settings")
	
	local uiElementName = guimgr:getElementByID("entrySettingsPlayername")
	local name = self:getPlayername()
	uiElementName:setText(name)
	
	local uiElementIP = guimgr:getElementByID("entrySettingsServerAddress")
	local ip = self:getServerAddress()
	uiElementIP:setText(ip)
	
end


function App:startGame()

	guimgr:changeUIstate("ingame")
	
	local players = lobby:getAllPlayers()
	local mapPath = lobby:getSelectedMapPath()
	lobby:destroy()
	
	-- inform clients if is host
	if self:isHost() then
		network:setupGame(mapPath, players)
	end
	
	game:create(mapPath, players)
	
end

-- no button callback! client calls this inside network code
function App:joinGame()
	guimgr:changeUIstate("ingame")
	
	local players = lobby:getAllPlayers()
	local mapPath = lobby:getSelectedMapPath()
	lobby:destroy()
	
	-- todo: confirm to server? confirm that map is ok (somewhere else)?
	--network:
	
	game:create(mapPath, players)
end


return App