--Mouse = require "scripts/mouse"
--GUIMgr	= require "scripts/guimgr"

local Button = {}

local textOffset = 6

-- button state:
-- 0 = disabled, not visible
-- 1 = active, normal
-- 2 = active, mouseover
-- 3 = active, mousedown
-- 4 = disabled (gray)

function Button:create(_desc)	--(_id, _x, _y, _textNormal, _callback)
	local et = {id = _desc.id, x = _desc.x or 0, y = _desc.y or 0,
		width = _desc.width or 128, height = _desc.height or 32}
	local button = entity:create(et)
	
	function button:load()
		gameloop:addLoop(self)
		renderer:addRenderer(self, LAYERGUI)
		self.id = _desc.id
		self.parent = _desc.parent
		self.image = imageloader.images["button1"][1]
		--self.height = 32
		--self.width = 128
		self.imageScaleX = self.width / self.image:getWidth()
		self.imageScaleY = self.height / self.image:getHeight()
		self.text = _desc.text
		self.textScale = 1.0
		self.visible = false
		self.state = 1
		self.hover = false
		self.active = false
		self.rect = {x = self.x, y = self.y, width = self.width, height = self.height}	--Rect:new(self.x, self.y, self.width, self.height)
		self.callback = function()
			userinput:resetFocus()
			if _desc.callback then
				_desc.callback()
			end
		end
		self.colors = { [0] = {1, 1, 1, 0}, [1] = {1, 1, 1, 1}, [2] = {0.5, 1, 1, 1}, [3] = {0.5, 1, 0.5, 1}, [4] = {0.8, 0.8, 0.8, 0.1} }
	end
	
	function button:tick(dt)
		--self.visible = (GUIMgr:getUIstate() == "main_menu")
		
		if not self.visible then
			return
		end
		
		if (self.state == 0) or (self.state == 4) then
			return
		end
		
		-- from here: execute only if the button is visible
		local mouseX, mouseY = love.mouse.getPosition()
		
		if self.active and (not love.mouse.isDown(1)) then
			if self.callback ~= nil then
				--soundmgr.playSound(soundmgr.sounds["game"]["bang2"])
				self.callback()
			end
		end
		
		if tools.isPointInsideRect(mouseX, mouseY, self.rect) then

			self.hover = true
			
			if love.mouse.isDown(1) then
				if tools.isPointInsideRect(mouse.pressedPos.x, mouse.pressedPos.y, self.rect) then
					self.active = true
				else
					self.active = false
				end
			else
				self.active = false
			end
			
		else
		
			self.hover = false
			self.active = false

		end
		
		if self.active then
			self.state = 3
		else
			if self.hover then
				self.state = 2
			else
				self.state = 1
			end
		end
		
	end
	
	--[[
	function button:tick(dt)
		--self.visible = (GUIMgr:getUIstate() == "main_menu")
		
		if not self.visible then
			return
		end
		
		-- from here: execute only if the button is visible
		local mouseX, mouseY = love.mouse.getPosition()
		if tools.isPointInsideRect(mouseX, mouseY, self.rect) then

			--if self.state ~= 4 then
			if (self.state ~= 4) and (tools.isPointInsideRect(Mouse.pressedPos.x, Mouse.pressedPos.y, self.rect)) then
			--if (self.state ~= 4) and (love.mouse.isDown(1)) then
				Mouse.uiElement = self.id
			end
			
			--self.textScale = 1.1
			if self.state ~= 3 and self.state ~= 4 then
				self.state = 2
			end
			
		else
		
			if Mouse.uiElement == self.id then
				Mouse.uiElement = nil
			end
			
			
			--self.textScale = 1.0
			if self.state == 2 or self.state == 3 then
				self.state = 1
			end

		end
		
	end
	--]]
	
	function button:draw()
		--[[
		if not self.visible then
			return
		end
		--]]
		
		
		-- drawable, x, y, rotation, scale x/y, offset x/y, shearing factor x/y
		love.graphics.draw(self.image, self.x, self.y, 0, self.imageScaleX, self.imageScaleY, 0, 0)
		
		r, g, b, a = love.graphics.getColor()
		love.graphics.setColor(self.colors[self.state])
		
		love.graphics.print(self.text, self.x + textOffset, self.y + textOffset, 0, self.textScale, self.textScale)
		
		love.graphics.setColor(r, g, b, a)
		
		--love.graphics.rectangle("line", self.x, self.y, self.width, self.height)
		
		--print("button get = " .. self.body:getX() .. ";" .. self.body:getY() .. "pos: " ..self.x .. ":" ..self.y)
	end
	
	return button
end


return Button