creature = require "scripts/entities/creature"

local Bat = {}

function Bat:create(_x, _y)

	local bat = creature:create({x = _x, y = _y})
	bat.entityType = "bat"
	
	function bat:load(_id)
		gameloop:addLoop(self)
		renderer:addRenderer(self, LAYERMOBS)
		
		self.id = _id
		self.images = imageloader.images["bat1"]
		self.size = 32
		self.imageHeight = self.images[1]:getHeight()
		self.imageWidth = self.images[1]:getWidth()
		self.imageScale = self.size / self.imageHeight
		self.movingVector = {x = 0, y = 0}	--{x = math.random(500), y = math.random(500)}	-- 2D directional vector
		self.speed = 0		-- current speed
		self.maxSpeed = 250	-- (world units = pixels?)
		self.rotation = 0
		self.maxAcceleration = 100
		self.inertnessFactor = 0.995	-- how fast the entity will slow down
		self.colliding = false
		self.movementType = "fly"
		self.behaviour = BEHAVIOUR_FOLLOW_PLAYER	--BEHAVIOUR_WANDER_AROUND
		self.rotateByMovement = false	-- do not rotate based on the current direction it faces
		self.followPlayerDistance = 500
		
		self.maxHealth = 50
		self.health = self.maxHealth
		self.displayHealthBar = true
		self.shield = 0
		self.maxShield = 0
		self.displayShieldBar = false
		
		self.animationTick = 0
		self.currentImageIndex = 1
		self.animationWait = 0.1
		
		-- abilities
		self.abilities["autoattack"] = {
			enabled = true,
			attackWait = 1.5,	-- value means it needs X seconds to attack once
			timer = 0,			-- set this to the same value as attackWait to allow the creature to use this ability immediately
			range = 10,
			damage = 5,
			condition = function() return self.currentTarget ~= nil end,
		}
		
	end
	
	
	function bat:tick(dt)
	
		self:move(dt)
		
		self:updateParticleSystems(dt)
		
		-- animation
		self.animationTick = self.animationTick + dt
		
		if self.animationTick >= self.animationWait then	-- * math.random(15, 25) / 10 then
			self:nextAnimation()
			self.animationTick = 0
		end
		
		
		-- abilities
		local ab_autoattack = self.abilities["autoattack"]
		if ab_autoattack ~= nil then				-- creature has this ability
			if ab_autoattack.enabled then			-- creature can use it (in general)
				ab_autoattack.timer = ab_autoattack.timer + dt
				
				if ab_autoattack.condition() then	-- condition fullfilled
					--print("creature: can autoattack")
					
					-- target in range? cooldown ok?
					local distanceSquared = tools.getDistanceSquared({x = self.x, y = self.y}, {x = self.currentTarget.x, y = self.currentTarget.y})
					-- measure from the targets AND the mobs outer face, not its center
					if (distanceSquared - (self.currentTarget.size/2) * (self.currentTarget.size/2)) <= ((ab_autoattack.range + self.size/2) * (ab_autoattack.range + self.size/2)) then
						if ab_autoattack.timer >= ab_autoattack.attackWait then
							ab_autoattack.timer = 0
							self:ab_autoAttackTarget(self.currentTarget)	-- use the ability!
						end
					end
				end
			end
		end
		
	end
	
	
	-- abilities ...
	function bat:ab_autoAttackTarget(target)
		local ab_autoattack = self.abilities["autoattack"]
		-- TODO: add some visuals and sound here
		--print("bat: hurt target")
		target:hurt(ab_autoattack.damage)
		self:recoil(0.6, self.maxAcceleration)	-- factor, [optional] minimum (total) recoil for both directions
	end
	
	
	function bat:nextAnimation()
		self.currentImageIndex = self.currentImageIndex + 1
		if self.images[self.currentImageIndex] == nil then
			self.currentImageIndex = 1
		end
	end
	
	
	--[[
	function bat:draw()
	
		local offsetX, offsetY = self.size / 2, self.size / 2
		love.graphics.draw(self.images[self.currentImageIndex], self.x + offsetX, self.y + offsetY, self.rotation, self.imageScale, self.imageScale, offsetX / self.imageScale, offsetY / self.imageScale)
		
		-- simple health bar
		if self.displayHealthBar then
			local health = self.health / self.maxHealth
			love.graphics.setColor(0, 0, 0, 0.9)
			--local height = self.health / 33
			local height = 6
			--love.graphics.rectangle("fill", self.x, self.y - height + 1, self.size, height)
			love.graphics.rectangle("fill", self.x, self.y - height + 1, self.size, height)
			local r, g, b = tools.getRGBbyHealth(health)
			love.graphics.setColor(r, g, b, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - height + 2, (self.size - 2) * health, height - 2)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
		-- simple shield bar
		if self.displayShieldBar and self.shield > 0 then
			local shield = self.shield / self.maxShield
			love.graphics.setColor(0, 0, 0, 0.9)
			love.graphics.rectangle("fill", self.x, self.y - 12, self.size, 6)
			love.graphics.setColor(0.8, 0.8, 0.8, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - 11, (self.size - 2) * shield, 4)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
	end
	--]]
	
	return bat
	
end


return Bat