projectile = require "scripts/entities/projectile"

local Bullet = {}

function Bullet:create(_x, _y, _movingVector, _sourceID, _size, _speed, _damage, _rotation)

	local bullet = projectile:create({x = _x, y = _y})
	bullet.entityType = "bullet"	-- no _ allowed!

	function bullet:load(_id)
		gameloop:addLoop(self)
		renderer:addRenderer(self, LAYEREFFECTS)
		--print("load bullet with id " .. _id)
		self.id = _id
		self.image = imageloader.images["player1"][1]
		self.size = _size
		self.imageHeight = self.image:getHeight()
		self.imageWidth = self.image:getWidth()
		self.imageScale = self.size / self.imageHeight
		self.movingVector = _movingVector	-- 2D directional vector
		self.speed = _speed		-- current speed
		self.damage = _damage
		self.rotation = _rotation or 0
		self.inertnessFactor = 1	-- how fast the entity will slow down; do not slow down at all
		self.colliding = false
		self.sourceID = _sourceID	-- who shot the bullet?

		-- center the bullet
		self.x = self.x - self.size/2
		self.y = self.y - self.size/2
	end
	
	
	function bullet:tick(dt)
	
		--print("bullet: tick")
		self:move(dt)
		
	end
	
	
	function bullet:move(dt)
		-- overriding...
		self.x = self.x + self.movingVector.x * dt
		self.y = self.y + self.movingVector.y * dt
	end
	
	
	function bullet:draw()
	
		--print("draw bullet")
		local offsetX, offsetY = self.size / 2, self.size / 2
		love.graphics.draw(self.image, self.x + offsetX, self.y + offsetY, self.rotation, self.imageScale, self.imageScale, offsetX / self.imageScale, offsetY / self.imageScale)
		
	end
	
	
	function bullet:destroy()
		--soundmgr.playSound(soundmgr.sounds["bang1"])
		
		-- copied from entity (mother "class")
		gameloop:removeLoop(self)
		renderer:removeRenderer(self)
		objmgr:destroyObject(self.id)
	end
	
	
	return bullet
	
end


return Bullet