-- moving entity <- entity

entity = require "scripts/entity"

local MovingEntity = {}

function MovingEntity:create(_entity)
	local movingEntity = entity:create({x = _entity.x, y = _entity.y})

	--print("creating moving entity named ".. _entity.id .." at " .. (_entity.x or 0) .. ";" .. (_entity.y or 0))
	
	movingEntity.movingVector		= {x = 0, y = 0}		-- 2D directional vector
	movingEntity.speed				= 0						-- current speed
	movingEntity.maxSpeed			= 200					-- (world units / pixels?)
	movingEntity.rotation			= 0
	movingEntity.maxAcceleration 	= 15
	movingEntity.inertnessFactor	= 0.9					-- how fast the movingEntity will slow down
	--movingEntity.movementType		= "walk"				-- movement type can be: walk, fly, swim
	--movingEntity.moving			= false					-- currently moving?
	movingEntity.rotateByMovement = true

	-- generic functions
	function movingEntity:load() end
	function movingEntity:tick(dt) end
	function movingEntity:draw() end

	
	-- get where the movingEntity is expected to be in the next step; top left corner
	function movingEntity:precalculateXandY(dt)
		return self.x + self.movingVector.x * dt, self.y + self.movingVector.y * dt
	end
	
	
	-- get where the entity is expected to be in the next step; center
	-- use dx and dy to give a different movingVector
	function movingEntity:precalculatePosition(dt, dx, dy)
		local offsetX, offsetY = (self.size or self.width) / 2, (self.size or self.height) / 2
		return self.x + (dx or self.movingVector.x) * dt + offsetX, self.y + (dy or self.movingVector.y) * dt + offsetY
	end
	
	
	function movingEntity:getMovingVector()
		return self.movingVector
	end
	
	
	function movingEntity:accelerate(dt, dx, dy, reset)
		local colliding = false
		
		if self.environment ~= nil then
			-- table that contains rectangles of blocking mapObjects around the moving entity
			for k, rect in pairs(self.environment) do
				local px, py = self:precalculatePosition(dt, dx, dy)
				local circle = {radius = self.size / 2, x = px, y = py}
				point = tools.getCollisionPoint(circle, rect)
				if point ~= nil then
					-- player is about to collide with the mapObject in the next step -> ignore the acceleration in this direction!
					--print("movingEntity colloding with: " .. rect.x .. ":" ..rect.y)
					colliding = true
					break
				end
			end
		end
		
		--print("moving entity accelerate - colliding:" .. tostring(colliding))
		
		if reset then
			if not colliding then	--and self.collisionPoint == nil
				self.movingVector.x = dx
				self.movingVector.y = dy
			end
		else
			if not colliding then	--and self.collisionPoint == nil
				self.movingVector.x = self.movingVector.x + dx
				self.movingVector.y = self.movingVector.y + dy
			end
		end
		
		--[[
		if not colliding then	--and self.collisionPoint == nil
			self.movingVector.x = self.movingVector.x + dx
			self.movingVector.y = self.movingVector.y + dy
		end
		--]]
		
	end
	
	-- sets the entitys position, careful with this!
	function movingEntity:setPosition(x, y)
		self.x = x
		self.y = y
	end
	
	
	function movingEntity:setRotation(_rot)
		self.rotation = _rot
	end
	
	
	function movingEntity:recoil(factor, minimum)
		local minValue = minimum or 0
		dx = self.movingVector.x * (-factor)
		dy = self.movingVector.y * (-factor)
		if math.abs(dx) < minValue then
			dx = tools.getSign(dx) * minValue
		end
		if math.abs(dy) < minValue then
			dy = tools.getSign(dy) * minValue
		end
		movingEntity:accelerate(AVG_DT, dx, dy, true)
	end
	
	function movingEntity:getPosition()
		--local offsetX, offsetY = self.imageWidth / 2, self.imageHeight / 2
		local offsetX, offsetY = self.size / 2, self.size / 2
		return self.x + offsetX, self.y + offsetY
	end
	
	-- returns table containing grid x and y index
	function movingEntity:getGridPosition()
		local _x, _y = self:getPosition()
		return {x = math.floor(_x/64), y = math.floor(_y/64)}
	end
	
	
	return movingEntity
end


return MovingEntity