creature = require "scripts/entities/creature"

local Player = {}

function Player:create(_x, _y, _humanPlayerId)
	local player = creature:create({x=_x, y=_y})
	player.entityType = "player"	-- no _ allowed!
	
	function player:load(_id)
		gameloop:addLoop(self)
		renderer:addRenderer(self, LAYERPLAYERS)
		
		self.id = _id	-- entity ID provided by the object manager
		--self.playerID = _id
		self.humanPlayerID = _humanPlayerId or 0
		--print("Player: created for human " .. self.humanPlayerID)
		self.image = imageloader.images["player2"][1]	--love.graphics.newImage("assets/graphics/player1.png")
		self.size = 48
		--self.height = self.image:getHeight()
		self.imageHeight = self.image:getHeight()
		--self.width = self.image:getWidth()
		self.imageWidth = self.image:getWidth()
		self.imageScale = self.size / self.imageHeight
		self.movingVector = {x = 0, y = 0}	-- 2D directional vector
		self.speed = 0		-- current speed
		self.maxSpeed = 250	-- (world units = pixels?)
		self.rotation = 0
		self.maxAcceleration = 20
		self.inertnessFactor = 0.95	-- how fast the entity will slow down
		self.colliding = false
		self.movementType = "walk"
		
		--self.partSys = {}
		self.hurtParticleImage = love.graphics.newImage("assets/graphics/player1.png")
		--self.hurtParticleTimer = 0
		
		self.runParticleImage = love.graphics.newImage("assets/graphics/player2.png")
		self.runParticleTimer = 0
		
		self.health = 200
		self.maxHealth = 200
		self.displayHealthBar = true
		self.shield = 0
		self.maxShield = 200
		self.displayShieldBar = true
		
		self.showCrosshairs = true
		self.rotateByMovement = false
		
		self.playerControlled = true	-- to not automatically use abilities; deprecated?
		
		self.bulletsToCreate = {}	-- not the best idea, but atm bullets will be created from inside the world
									-- this list simply tells the world what bullets should be created
		
		-- abilities
		self.abilities["healthregeneration"] = {
			enabled = true,
			intervall = 1,		-- value means it needs X seconds before the ability can fire again
			timer = 0,			-- set this to the same value as intervall to allow the creature to use this ability immediately
			heal = 5,
			condition = function() return (self.health < self.maxHealth) and (self.lastHurtTime + 2 < GAMETIME) end,
		}
		
		self.abilities["rangedattack"] = {
			enabled = true,
			attackWait = 0.5,	-- must be > 0
			timer = 0,
			damage = 25,
			bulletSpeed = 800,
			bulletSize = 10,
			condition = function() return (love.mouse.isDown(1) == true) and guimgr:getUIstate() == "ingame" end,
		}

	end
	
	function player:tick(dt)
	
		--print("player rotation: " .. self.rotation)

		local offsetX, offsetY = self.size / 2
		local offsetX, offsetY = self.imageWidth / 2, self.imageHeight / 2--
		
		local human = app:getHumanPlayer()
		
		if human.id == self.humanPlayerID then
			camera:setPosition(self.x + offsetX, self.y + offsetY)
			
			-- if necessary: rotate the player by the mouse aim
			if not self.rotateByMovement then
				local mx, my = mouse:getWorldPosition()
				local px, py = self:getPosition()
				self.rotation = math.atan2((my - py), (mx - px)) + math.pi * 0.5
			end
		end
		
		-- moved to userinput
		--[[
		if not self.colliding then
			if love.keyboard.isDown("a") then
				--self.movingVector.x = self.movingVector.x - self.maxAcceleration
				self:accelerate(dt, - self.maxAcceleration, 0)
				
			elseif love.keyboard.isDown("d") then
				--self.movingVector.x = self.movingVector.x + self.maxAcceleration
				self:accelerate(dt, self.maxAcceleration, 0)
				
			end
			
			if love.keyboard.isDown("w") then
				--self.movingVector.y = self.movingVector.y - self.maxAcceleration#
				self:accelerate(dt, 0, - self.maxAcceleration)
				
			elseif love.keyboard.isDown("s") then
				--print("player: move down")
				--self.movingVector.y = self.movingVector.y + self.maxAcceleration
				self:accelerate(dt, 0, self.maxAcceleration)
				
			end
		end
		--]]
		
		if self.moving then
			self.runParticleTimer = self.runParticleTimer + dt
		end
		
		-- move the player
		self:move(dt)
		
		if (self.runParticleTimer > 0.2) and (self.speed > 100) then
			self:createRunningParticles()
			self.runParticleTimer = 0
		end
		
		
		self:updateParticleSystems(dt)
		
		--[[
		-- update particle systems
		local numberOfParticleSystems = #self.partSys
		if numberOfParticleSystems > 0 then
			for i = numberOfParticleSystems, 1, -1 do
				local pMin, pMax = self.partSys[i].system:getParticleLifetime()
				local lifetime = self.partSys[i].system:getEmitterLifetime()
				-- also take the particle lifetime into consideration, not only the emitter's!
				-- emitter lifetime of -1 means it emitts forever -> do not remove
				if (self.partSys[i].startTime + lifetime + pMax <= GAMETIME) and (lifetime > -1) then
					--print("ps: remove")
					table.remove(self.partSys, i)
				else
					--if self.partSys[i].name == "run" then
						--local x,y = self.partSys[i].system:getPosition()
						--print(x,y)
						--self.partSys[i].system:setPosition(self.x + offsetX, self.y + offsetY)
						--self.partSys[i].pos = {X = self.x + offsetX, Y = self.y + offsetY}
					--end
					self.partSys[i].system:update(dt)
				end
			end
		end
		--]]
		
		
		-- abilities
		local ab_healthregeneration = self.abilities["healthregeneration"]
		if ab_healthregeneration ~= nil then				-- creature has this ability
			if ab_healthregeneration.enabled then			-- creature can use it (in general)
				ab_healthregeneration.timer = ab_healthregeneration.timer + dt
				if ab_healthregeneration.condition() then	-- condition fullfilled
					-- cooldown ok?
					if ab_healthregeneration.timer >= ab_healthregeneration.intervall then
						ab_healthregeneration.timer = 0
						self:ab_regenerateHealth()	-- use the ability!
					end
				end
			end
		end
		
		local ab_rangedattack = self.abilities["rangedattack"]
		if ab_rangedattack ~= nil then				-- creature has this ability
			if ab_rangedattack.enabled then			-- creature can use it (in general)
				ab_rangedattack.timer = ab_rangedattack.timer + dt
				if ab_rangedattack.condition() then	-- condition fullfilled
					-- cooldown ok?
					if ab_rangedattack.timer >= ab_rangedattack.attackWait then
						ab_rangedattack.timer = 0
						self:ab_doRangedAttack()	-- use the ability!
					end
				end
			end
		end
		
		
	end	-- end:tick
	
	
	-- abilities ...
	function player:ab_regenerateHealth()
		local ab_healthregeneration = self.abilities["healthregeneration"]
		-- TODO: add some visuals and sound here
		--print("player: regenerate health")
		self:heal(ab_healthregeneration.heal)
	end
	
	
	function player:ab_doRangedAttack()
		--print("player: fire bullet")
		local ab_rangedattack = self.abilities["rangedattack"]
		local bullet = {}
		local px, py = self:getPosition()
		bullet.x = px
		bullet.y = py
		bullet.movingVector = tools.setVectorLength(tools.getVectorFromAngle(self.rotation - math.pi * 0.5), ab_rangedattack.bulletSpeed)
		bullet.damage = ab_rangedattack.damage
		bullet.speed = ab_rangedattack.bulletSpeed
		bullet.size = ab_rangedattack.bulletSize
		bullet.rotation = self.rotation
		bullet.sourceID = self.id
		table.insert(self.bulletsToCreate, bullet)
		
		-- right place? probably not
		soundmgr.playSound(soundmgr.sounds["game"]["bang1"])
	end
	
	
	-- note: rotation uses the anchor which is the top left corner by default
	-- this changes by giving a different offset which in turn changes the position where the image is drawn
	-- because drawing still starts from the top left corner
	-- camera goto_point must also be adjusted with the new anchor
	-- offset x and y in graphics.draw function must be divided by the imageScale
	function player:draw()
		local human = app:getHumanPlayer()
		
		-- TODO: this should not be inside the player? (for the multiplayer this would not work...)
		if guimgr:getUIstate() ~= "ingame" and guimgr:getUIstate() ~= "ingame_settings" then
			return
		end
		
		if self.showCrosshairs and human.id == self.humanPlayerID then
			local mx, my = mouse:getWorldPosition()
			local px, py = self:getPosition()
			love.graphics.setColor(180, 180, 180, 0.2)
			love.graphics.line(px, py, mx, my)
			love.graphics.setColor(0, 0, 0, 0.4)
			love.graphics.setLineWidth(2)
			love.graphics.circle("line", mx, my, 15)
			love.graphics.line(mx + 17, my, mx - 17, my)
			love.graphics.line(mx, my + 17, mx, my - 17)
			love.graphics.setColor(255, 255, 255, 255)
			love.graphics.setLineWidth(1)
		end
		
		-- drawable, x, y, rotation, scale x/y, offset x/y, shearing factor x/y
		local offsetX, offsetY = self.size / 2, self.size / 2
		love.graphics.draw(self.image, self.x + offsetX, self.y + offsetY, self.rotation, self.imageScale, self.imageScale, offsetX / self.imageScale, offsetY / self.imageScale)
		--love.graphics.rectangle("line", self.x, self.y, self.size, self.size)
		
		
		-- testing... collision
		--[[
		if self.collisionLine ~= nil then
			love.graphics.line(self.collisionLine[1].x, self.collisionLine[1].y, self.collisionLine[2].x, self.collisionLine[2].y)
		end
		--]]
		if self.collisionPoint ~= nil then
			love.graphics.circle("fill", self.collisionPoint.x, self.collisionPoint.y, 3)
		end
		
		
		-- simple health bar
		if self.displayHealthBar then
			local health = self.health / self.maxHealth
			love.graphics.setColor(0, 0, 0, 0.9)
			love.graphics.rectangle("fill", self.x, self.y - 5, self.size, 6)
			local r, g, b = tools.getRGBbyHealth(health)
			love.graphics.setColor(r, g, b, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - 4, (self.size - 2) * health, 4)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
		-- simple shield bar
		if self.displayShieldBar and self.shield > 0 then
			local shield = self.shield / self.maxShield
			love.graphics.setColor(0, 0, 0, 0.9)
			love.graphics.rectangle("fill", self.x, self.y - 12, self.size, 6)
			love.graphics.setColor(0.8, 0.8, 0.8, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - 11, (self.size - 2) * shield, 4)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
		
		-- particle system
		local numberOfParticleSystems = #self.partSys
		if numberOfParticleSystems > 0 then
			for i = 1, numberOfParticleSystems do
				--local x,y = self.partSys[i].system:getPosition()
				love.graphics.draw(self.partSys[i].system, self.partSys[i].pos.X, self.partSys[i].pos.Y)
			end
		end
		
	end

	
	function player:getXandY()
		return self.x, self.y
	end
	
	function player:getRotation()
		return self.rotation
	end
	
	
	-- moved to movingEntity
	--[[
	function player:getPosition()
		--local offsetX, offsetY = self.imageWidth / 2, self.imageHeight / 2
		local offsetX, offsetY = self.size / 2, self.size / 2
		return self.x + offsetX, self.y + offsetY
	end
	--]]
	
	
	function player:heal(_heal)
		-- TODO: add particles
		
		self.health = math.min(self.maxHealth, self.health + _heal)	-- note that there is no check if _heal is > 0
	end
	
	--[[
	function player:hurt(_damage, _ignoreshield)
		local offsetX, offsetY = self.imageWidth / 2, self.imageHeight / 2
		
		-- 
		local damage = _damage or 1
		local ignoreshield = _ignoreshield or false
		local damageHealth, damageShield = 0, 0
		if self.shield == 0 or ignoreshield then
			-- player has no shield OR takes true damage -> take full damage to healthbar
			damageHealth = damage
		else
			-- player has shield, damage shield first and raw health if necessary
			damageShield = math.min(damage, self.shield)
			damageHealth = damage - damageShield
		end
		self.shield = math.max(0, self.shield - damageShield)
		self.health = math.max(0, self.health - damageHealth)
		--print("new health: " .. self.health)
		
		
		-- TODO: display particle color based on shield / raw health loss - adjust colors!
		-- disply particles
		local r, g, b = 0, 0, 255
		if GAMETIME - self.hurtParticleTimer > 0.5 then
			local sys = {}
			if damageHealth > 0 then
				r, g, b = 255, 0, 0
			end
			sys.system = love.graphics.newParticleSystem(self.hurtParticleImage, 64)
			sys.system:setParticleLifetime(0.5, 1) -- Particles live at least (min)s and at most (max)s.
			sys.system:setEmissionRate(64)
			sys.system:setSizeVariation(1)
			sys.system:setEmitterLifetime(0.1)
			sys.system:setEmissionArea("uniform", self.size / 2, self.size / 2)
			sys.system:setLinearAcceleration(-20, -20, 20, 20) -- Random movement in all directions.
			sys.system:setColors(r, g, b, 255, r, g, b, 0) -- Fade to transparency.
			sys.system:setSizes(0.1, 0.2)
			sys.pos = {X = self.x + offsetX, Y = self.y + offsetY}
			sys.startTime = GAMETIME
			sys.name = ""
			table.insert(self.partSys, sys)
			
			self.hurtParticleTimer = GAMETIME
		end
		
		self.lastHurtTime = GAMETIME
		
	end
	--]]
	
	
	function player:createRunningParticles()
		-- show running particles
		local offsetX, offsetY = self.imageWidth / 2, self.imageHeight / 2
		local sys = {}
		sys.system = love.graphics.newParticleSystem(self.runParticleImage, 32)
		sys.system:setParticleLifetime(0.5, 1) -- Particles live at least (min)s and at most (max)s.
		sys.system:setEmissionRate(8)
		sys.system:setSizeVariation(1)
		sys.system:setEmitterLifetime(0.5)
		sys.system:setEmissionArea("uniform", self.size / 2, self.size / 2)
		sys.system:setLinearAcceleration(-20, -20, 20, 20) -- Random movement in all directions.
		sys.system:setColors(255, 255, 255, 255, 255, 255, 255, 0) -- Fade to transparency.
		sys.system:setSizes(0.1)
		sys.pos = {X = self.x + offsetX, Y = self.y + offsetY}
		sys.startTime = GAMETIME
		sys.name = "run"
		--sys.system:setPosition(self.x + offsetX, self.y + offsetY)
		table.insert(self.partSys, sys)
	end
	
	return player
end


return Player