creature = require "scripts/entities/creature"
--soundmgr	= require "scripts/soundmgr"

local Zombie = {}

function Zombie:create(_x, _y)

	local zombie = creature:create({x = _x, y = _y})
	zombie.entityType = "zombie"
	
	function zombie:load(_id)
		gameloop:addLoop(self)
		renderer:addRenderer(self, LAYERMOBS)
		
		self.id = _id
		self.images = imageloader.images["zombie1"]
		self.image = imageloader.images["zombie1"]["up"][1]
		self.size = 48
		self.imageHeight = self.image:getHeight()
		self.imageWidth = self.image:getWidth()
		self.imageScale = self.size / self.imageHeight
		
		self.movingVector = {x = 0, y = 0}	--{x = math.random(500), y = math.random(500)}	-- 2D directional vector
		self.speed = 0		-- current speed
		self.maxSpeed = 120	-- (world units = pixels?)
		self.rotation = 0
		self.maxAcceleration = 60
		self.inertnessFactor = 0.995	-- how fast the entity will slow down
		self.colliding = false
		self.movementType = "walk"
		self.behaviour = BEHAVIOUR_FOLLOW_PLAYER
		self.followPlayerDistance = 300
		
		self.maxHealth = math.random(1, 4) * 100
		self.health = self.maxHealth
		self.displayHealthBar = true
		self.shield = math.random(1, 4) * 20		-- testing...
		self.maxShield = self.shield
		self.displayShieldBar = true
		
		self.animationTick = 0
		self.currentImageIndex = 1
		self.animationWait = 0.2
		self.animationSet = "down"
		
		self.sounds = {}
		self.sounds["hurt"] = {}
		self.sounds["hurt"][1] = soundmgr.sounds["game"]["zombie1_misc1"]--; self.sounds["hurt"][1]:setVolume(0.7)
		self.sounds["hurt"][2] = soundmgr.sounds["game"]["zombie1_misc2"]--; self.sounds["hurt"][2]:setVolume(0.3)
		
		-- abilities
		self.abilities["autoattack"] = {
			enabled = true,
			attackWait = 2.5,	-- value means it needs X seconds to attack once
			timer = 0,			-- set this to the same value as attackWait to allow the creature to use this ability immediately
			range = 20,
			damage = 30,
			condition = function()
				if self.currentTargetID == nil then
					return false
				else
					if objmgr:getObjectByID(self.currentTargetID) == nil then
						self:setAttackTarget(nil)
						return false
					else
						return true
					end
				end
			end,
		}
		
	end
	
	
	function zombie:tick(dt)
	
		self:move(dt)
		
		self:updateParticleSystems(dt)
		
		
		-- "animate" the zombie + change the set of images depending on its rotation
		self:updateAnimation(dt)
		
		
		-- abilities
		local ab_autoattack = self.abilities["autoattack"]
		if ab_autoattack ~= nil then				-- creature has this ability
			if ab_autoattack.enabled then			-- creature can use it (in general)
				ab_autoattack.timer = ab_autoattack.timer + dt
				
				if ab_autoattack.condition() then	-- condition fullfilled
					--print("creature: can autoattack")
					
					-- target in range? cooldown ok?
					local currentTarget = objmgr:getObjectByID(self.currentTargetID)
					--if currentTarget == nil then
						--self:setAttackTarget(nil)
					--end
					local tarPosX, tarPosY = currentTarget:getPosition()
					local distanceSquared = tools.getDistanceSquared({x = self.x, y = self.y}, {x = tarPosX, y = tarPosY})
					-- measure from the targets AND the mobs outer face, not its center
					if (distanceSquared - (currentTarget.size/2) * (currentTarget.size/2)) <= ((ab_autoattack.range + self.size/2) * (ab_autoattack.range + self.size/2)) then
						if ab_autoattack.timer >= ab_autoattack.attackWait then
							ab_autoattack.timer = 0
							self:ab_autoAttackTarget(self.currentTargetID)	-- use the ability!
						end
					end
				end
			end
		end
		
	end
	
	
	function zombie:updateAnimation(dt)
		self.animationTick = self.animationTick + dt
		
		if self.animationTick >= self.animationWait then	-- * math.random(15, 25) / 10 then
			self:nextAnimation()
			self.animationTick = 0
		end
		
		local r = self.rotation
		if (r >= math.pi * 1.25) or (r < -math.pi * 0.25) then
			self.animationSet = "left"
		elseif (r < math.pi * 0.25) then
			self.animationSet = "up"
		elseif (r < math.pi * 0.75) then
			self.animationSet = "right"
		elseif (r < math.pi * 1.25) then
			self.animationSet = "down"
		end
		
		self.image = self.images[self.animationSet][self.currentImageIndex]
		
	end
	
	
	function zombie:nextAnimation()
		self.currentImageIndex = self.currentImageIndex + 1
		if self.images[self.animationSet][self.currentImageIndex] == nil then
			self.currentImageIndex = 1
		end
	end
	
	
	-- abilities ...
	function zombie:ab_autoAttackTarget(targetID)
		local ab_autoattack = self.abilities["autoattack"]
		local target = objmgr:getObjectByID(targetID)
		-- TODO: add some visuals and sound here
		target:hurt(ab_autoattack.damage, self.id)
		self:recoil(0.6, self.maxAcceleration)	-- factor, [optional] minimum (total) recoil for both directions
		soundmgr.playSound(soundmgr.sounds["game"]["roar1"])
	end
	
	
	--
	function zombie:draw()
	
		local offsetX, offsetY = self.size / 2, self.size / 2
		love.graphics.draw(self.image, self.x + offsetX, self.y + offsetY, 0, self.imageScale, self.imageScale, offsetX / self.imageScale, offsetY / self.imageScale)
		
		-- simple health bar
		if self.displayHealthBar then
			local health = self.health / self.maxHealth
			love.graphics.setColor(0, 0, 0, 0.9)
			--local height = self.health / 33
			local height = 6
			--love.graphics.rectangle("fill", self.x, self.y - height + 1, self.size, height)
			love.graphics.rectangle("fill", self.x, self.y - height + 1, self.size, height)
			local r, g, b = tools.getRGBbyHealth(health)
			love.graphics.setColor(r, g, b, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - height + 2, (self.size - 2) * health, height - 2)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
		-- simple shield bar
		if self.displayShieldBar and self.shield > 0 then
			local shield = self.shield / self.maxShield
			love.graphics.setColor(0, 0, 0, 0.9)
			love.graphics.rectangle("fill", self.x, self.y - 12, self.size, 6)
			love.graphics.setColor(0.8, 0.8, 0.8, 0.9)
			love.graphics.rectangle("fill", self.x + 1, self.y - 11, (self.size - 2) * shield, 4)
			-- reset drawing color
			love.graphics.setColor(255, 255, 255, 255)
		end
		
		-- particle systems
		local numberOfParticleSystems = #self.partSys
		if numberOfParticleSystems > 0 then
			for i = 1, numberOfParticleSystems do
				--local x,y = self.partSys[i].system:getPosition()
				love.graphics.draw(self.partSys[i].system, self.partSys[i].pos.X, self.partSys[i].pos.Y)
			end
		end
		
	end
	--
	
	
	return zombie
	
end


return Zombie