world = require "scripts/world"

-- the actual game

local Game = {}
local APPTYPE_SOLO = 0
local APPTYPE_SERVER = 1
local APPTYPE_CLIENT = 2


function Game:init()
	
	self.players = {}
	
	local btnBackToGame = {
		parent = "ingame_settings",
		id = "btnIngameSettingsToGame",
		x = 128,
		y = 64,
		text = "Back to Game",
		callback = function() app:backToGame() end,
	}
	
	local btnMainMenu = {
		parent = "ingame_settings",
		id = "btnIngameSettingsToMainMenu",
		x = 128,
		y = 272,
		text = "Quit to Main Menu",
		callback = function() app:backToMainMenu() end,
	}
	guimgr:add(button:create(btnBackToGame))
	guimgr:add(button:create(btnMainMenu))
	
end

-- passed in table contains all players from the lobby
function Game:create(_mapPath, _playerTable)

	-- start a new game
	
	local playerCount = table.getn(_playerTable)
	--print("players: "..playerCount)
	
	
	-- which map to load? _mapPath
	local _world = world:create(_mapPath)
	objmgr:addObject(_world)
	self:setWorld(_world)
	
	
	for p = 1, playerCount do
		local newPlayer = {}
		newPlayer.humanID = _playerTable[p].id
		newPlayer.displayname = _playerTable[p].displayname
		newPlayer.connection = _playerTable[p].connection
		local spawnPosX, spawnPosY = self.world:getPlayerSpawnPosition(newPlayer.humanID)
		local entityID
		
		if app:getType() ~= APPTYPE_CLIENT then	-- do this for singleplayer and host; not for clients
			_player = player:create(64 * (spawnPosX - 1) + 8, 64 * (spawnPosY - 1) + 8, newPlayer.humanID)	-- added + 8 to center the player
			entityID = objmgr:addObject(_player)
			
			newPlayer.objID = entityID
			self.world:addPlayer(newPlayer.humanID, entityID)		-- game player object (human player ID, entityID)
		end
		
		self:addPlayer(newPlayer)		-- human player
		
		-- server: send info to all clients
		-- TODO: entity type? make different functions for different entity types
		if app:getType() == APPTYPE_SERVER then
			network.server.broadcast("newply:" .. 64 * (spawnPosX - 1) + 8 .. ":" .. 64 * (spawnPosY - 1) + 8 .. ":" .. newPlayer.humanID .. ":" .. entityID)
		end
		
	end
	
end

-- information received from the server as client
function Game:updatePlayerEntity(_entityTable)
	local newPlayer = { humanID = _entityTable.humanID, objID = _entityTable.entityID }	-- human
	local _player = player:create(_entityTable.x, _entityTable.y, _entityTable.humanID)	-- game entity
	local entityID = objmgr:addObject(_player, _entityTable.entityID)
	
	-- entityID and _entityTable.entityID should be the same (see objmgr)
	self:updatePlayerEntityID(_entityTable.humanID, _entityTable.entityID)		-- human player
	self.world:addPlayer(_entityTable.humanID, _entityTable.entityID)			-- game player object (human player ID, entityID)
end

-- TODO: any entity (not only player!)
-- information received from the server as client
function Game:updateEntity(_entityTable)
	local newPlayer = { humanID = _entityTable.humanID, objID = _entityTable.entityID }	-- human; TODO (above!)
	local _player = player:create(_entityTable.x, _entityTable.y, _entityTable.humanID)	-- game entity
	local entityID = objmgr:addObject(_player, _entityTable.entityID)
	--print("Game: objID (1) " .. _entityTable.entityID)
	--print("Game: objID (2) " .. entityID)
	
	-- entityID and _entityTable.entityID should be the same (see objmgr)
	self:updatePlayerEntityID(_entityTable.humanID, _entityTable.entityID)		-- human player
	self.world:addPlayer(_entityTable.humanID, _entityTable.entityID)			-- game player object (human player ID, entityID)
end


function Game:updatePlayerEntityID(_humanID, _entityID)
	--print("Game: try update player entity id for human: " .. _humanID .. " -> " .. _entityID)
	local player = self:getPlayerByHumanID(_humanID)
	if player then
		player.objID = _entityID
		return true
	end
	return false
end


function Game:getPlayerEntityByHumanID(_humanID)
	--return self.world:getPlayerEntityByHumanID(_humanID)
	
	for i = 1, table.getn(self.players) do
		local ply = self.players[i]
		if ply.humanID == _humanID then
			return ply.objID
		end
	end
	
	return nil
end


function Game:setWorld(_world)
	self.world = _world
end


-- note that this is a human player
function Game:addPlayer(_player)

	--[[
	print(" ")
	print("id: " .. _player.id)
	print("name: " .. _player.displayname)
	print("conn: " .. _player.connection)
	--]]
	
	local player = {}
	player.humanID = _player.humanID	-- human ID
	player.displayname = _player.displayname or "Spieler " .. player.humanID
	player.connection = _player.connection or 0
	player.objID = _player.objID
	
	table.insert(self.players, player)
	
	print("Game: there are currently " .. #self.players .. " players in the game")
	
	return player
	
end


function Game:removePlayerByHumanId(_humanID)
	for i = #self.players, 1, -1 do
		local player = self.players[i]
		if player.humanID == _humanID then
			table.remove(self.players, i)
			return true
		end
	end
	return false
end


function Game:removePlayer(_player)	-- TODO: now remove by table; should be by (entity or human player) ID

	for i = #self.players, 1, -1 do
		local player = self.players[i]
		if player.objID == _player.objID then
			table.remove(self.players, i)
			return true
		end
	end
	return false
end

function Game:getPlayerByHumanID(_humanID)

	for i = 1, #self.players do
		local ply = self.players[i]
		if ply.humanID == _humanID then
			return ply
		end
	end

	return nil
end

function Game:getPlayerByEntityID(_objID)

	for i = 1, #self.players do
		local ply = self.players[i]
		if ply.objID == _objID then
			return ply
		end
	end

	return nil
end


function Game:destroy()

	-- remove all players etc
	for i = #self.players, 1, -1 do
		local player = self.players[i]
		self:removePlayer(player)
		
		--objmgr:getObjectByID(player.objID):destroy()	-- remove?
		
		self.world:destroy()
		
		table.remove(self.players, i)
	end
	
end


return Game

