
-- the lobby

local Lobby = {}
local APPTYPE_SOLO = 0
local APPTYPE_SERVER = 1
local APPTYPE_CLIENT = 2

function Lobby:load()

	self.players = {}
	--self.visible = true	-- in order to render everything it must be declared visible
	self.selectedMapFile = nil
	
	gameloop:addLoop(self)
	renderer:addRenderer(self, LAYERGUI)
	
	local btnStart = {
		parent = "lobby",
		id = "btnLobbyStartGame",
		x = 128,
		y = 300,
		text = "Start Game",
		callback = function() app:startGame() end,
	}
	
	local btnBack = {
		parent = "lobby",
		id = "btnLobbyToMainMenu",
		x = 280,
		y = 300,
		text = "Back to Main Menu",
		callback = function() app:backToMainMenu() end,
	}
	
	guimgr:add(button:create(btnStart))
	guimgr:add(button:create(btnBack))
	
	-- maps
	guimgr:add(label:create({parent = "lobby", id = "lblLobbyMapsTitle", text = "MAPS", x = 340, y = 100}))
	
	-- TODO: add some validation...
	local files = love.filesystem.getDirectoryItems("/assets/maps")
	for k, file in ipairs(files) do
		--print("lobby: " .. k .. ". " .. file) --outputs something like "1. map1.lua"
		
		if string.find(file, ".lua") ~= nil then
			--print("lobby: " .. k .. ". " .. file)
			if self.mapFiles == nil then
				self.mapFiles = {}
				self.selectedMapFile = string.sub(file, 1, -5)
			end
			table.insert(self.mapFiles, file)
			
			-- test: create a button for every map
			local index = #self.mapFiles
			local fileName = string.sub(file, 1, -5)
			local btnMap = {
				parent = "lobby",
				id = "btnLobbySelectMap_"..index,
				x = 320,
				y = 100 + index * 40,
				text = fileName,
				callback = function()
					-- host?
					local appType = app:getType()
					--local human = app:getHumanPlayer()
					if appType == APPTYPE_SOLO then
						self:selectMapFile(fileName)
						--print("Lobby: select map: " .. fileName)
					
					elseif appType == APPTYPE_SERVER then
						self:selectMapFile(fileName)
						network.server.sendMapInfo(fileName)
						--print("Lobby: select map: " .. fileName)
						
					else	-- client
						print("Lobby: client is not allowed to change map!")
					end
				end,
			}
			guimgr:add(button:create(btnMap))
			
		end
	end

	guimgr:add(label:create({parent = "lobby", id = "lblLobbyPlayerTitle", text = "PLAYERS", x = 100, y = 100}))
	--[[
	for i = 1, #self.players do
		local player = self.players[i]
		guimgr:add(label:create({parent = "lobby", id = "lblLobbyPlayer_" .. i, text = player.displayname, x = 200, y = 100 + 40 * i}))
	end
	--]]
	
end


function Lobby:selectMapFile(fileName)
	self.selectedMapFile = fileName
	print("Lobby: select map: " .. fileName)
end


function Lobby:create()
	self.players = {}
	gameloop:addLoop(self)
	renderer:addRenderer(self, LAYERGUI)
	
end


function Lobby:getSelectedMapPath()
	return "assets/maps/" .. self.selectedMapFile	-- no ".lua" here!
end

function Lobby:tick(dt)
	if guimgr:getUIstate() ~= "lobby" then
		return
	end
	--print("lobby: tick")
	
	--[[
	if network then
		local conn = network:getConnections()
		local numConnections = #conn or 0
		if numConnections > 0 then
			--print("lobby: there are " .. numConnections .. " connections")
		end
	end
	--]]
	
	--self:sortPlayers()
end


function Lobby:draw()

	if guimgr:getUIstate() ~= "lobby" then
		return
	end
	
	-- local player
	for i = 1, #self.players do
		local player = self:getPlayerByID(i)	--self.players[i]
		--love.graphics.print("player " .. i .. ": " .. player.displayname, 100, 100 + 40 * i, 0, 1, 1)
		if player == nil then
			love.graphics.print("??? " .. i, 100, 100 + 40 * i, 0, 1, 1)
		else
			local dispname = player.displayname or "?"
			love.graphics.print("[Player " .. player.id .. "]: " .. dispname .. " at " .. player.ip .. " (" .. player.port .. ")", 100, 100 + 40 * i, 0, 1, 1)
		end
	end
	
	-- network players
	--[[
	if network then
		local connections = network:getConnections()
		local numConnections = #connections or 0
		if numConnections > 0 then
			for i = 1, numConnections do
				local conn = connections[i]
				--print("lobby: there are " .. numConnections .. " connections")
				local dispname = conn.displayname or "?"
				love.graphics.print("player " .. i .. ": " .. dispname .. " at " .. conn.ip .. " (" .. conn.port .. ")", 100, 150 + 40 * i, 0, 1, 1)
			end
		end
	end
	--]]
	
	local mapText = "(none)"
	if self.selectedMapFile ~= nil then
		mapText = self.selectedMapFile
	end
	love.graphics.print("Selected Map: " .. mapText, 440, 100)
	
	--love.graphics.print("||== maps ==||", 340, 100)
	--for j = 1, #self.mapFiles do
		--local fileName = string.sub(self.mapFiles[j], 1, -5)
		--love.graphics.print(j .. ". map: " .. fileName, 320, 100 + 40 * j)
	--end
	
end


function Lobby:destroy()
	gameloop:removeLoop(self)
	renderer:removeRenderer(self)
end


-- note that this is a human player
function Lobby:addPlayer(_player)

	--if Lobby:getPlayerByID(_player.id) == nil then
	
	if _player.id and self:getPlayerByID(_player.id) ~= nil then
		print("Lobby: Error - there is already a player with ID " .. _player.id)
		return false
	end
	
	local player = {}
	player.id = _player.id or 0
	player.displayname = _player.displayname	-- or "Spieler " .. player.id
	player.ip = _player.ip or 0
	player.port = _player.port or 0
	
	table.insert(self.players, player)
	
	print("Lobby: Add player with ID " .. _player.id)
	print("Lobby: There are currently " .. #self.players .. " players in the lobby")
	
	self:sortPlayers()
	
	return player
	
end

-- TODO: index = id ??? or just a key ???
function Lobby:removePlayerById(_id)
	if self.players[_id] then
		print("Lobby: remove player with Id " .. _id)
		self.players[_id] = nil
		return true
	else
		print("Lobby: CANNOT remove player with Id " .. _id)
		return false
	end
end


function Lobby:printPlayers()	-- debug print only!
	for i = 1, #self.players do
		local player = self.players[i]
		print("Lobby: player " .. i .. " id: " .. player.id .. " " .. player.displayname)
	end
end


function Lobby:updatePlayer(_player)
	if _player.id and self:getPlayerByID(_player.id) == nil then
		print("Lobby: Error - there is no player to update with ID " .. _player.id)
		return false
	end
	
	-- TODO: ?
	local player = self:getPlayerByID(_player.id)
	--if _player.id ~= nil then
	
	-- sort player table
	self:sortPlayers()
end

-- not used because error: compare number with string (???)
function Lobby:sortPlayers()
	if #self.players < 2 then return false end
	
	for i = #self.players, 2, -1 do
		local player = self.players[i]
		local previousPlayer = self.players[i-1]
		--[[
		print(player.id)
		print(previousPlayer.id)
		if type(player.id) ~= "number"
		or type(previousPlayer.id) ~= "number" then
			return false
		end
		--]]
		if (previousPlayer.id) > (player.id) then
			-- swap!
			local change = player
			player = previousPlayer
			previousPlayer = change
			-- test...
			--self.sortPlayers = function() end
		end
	end
	
	return true
end


function Lobby:setPlayerName(_connection)
	for i = #self.players, 1, -1 do
		local player = self.players[i]
		if player.ip == _connection.ip
		and player.port == _connection.port then
			player.displayname = _connection.displayname
			break
		end
	end
end

function Lobby:removePlayer(_player)

	for i = #self.players, 1, -1 do
		local player = self.players[i]
		if player.id == _player.id then
			table.remove(self.players, i)
			break
		end
	end
	
end

function Lobby:getPlayerByID(_id)

	for i = 1, #self.players do
		local ply = self.players[i]
		--print("Lobby: searching player with id " .. _id .. " -  " .. i .. " id: " .. ply.id)
		if ply.id == _id then
			--print("Lobby: " .. _id .. " = " .. ply.id)
			return ply
		end
		-- prints 2 != 2 on client when not using tonumber (wtf)
		--print("Lobby: " .. _id .. " != " .. ply.id)
	end

	return nil
end

function Lobby:getAllPlayers()
	self:sortPlayers()	--?
	return self.players
end

return Lobby

