terrain		= require "scripts/terrain"
block		= require "scripts/mapobjects/block"
stone		= require "scripts/mapobjects/stone"
grass		= require "scripts/mapobjects/grass"
water		= require "scripts/mapobjects/water"
a_star		= require "scripts/a_star"

-- TODO: build maps in a way that there is one big background (images only, no mapObjects)
-- ... and put zeros where there are no mapObjects, the other mapObjects are created similar to now
-- ... the advantage will be the visuals: easy transitions between different terrains

local Map = {}

function Map:create(_mapPath)

	local map = {}
	
	--map.size = {height = 64, width = 64}
	--map.data = {}	-- contains map = level informaton (environment)
	
	map = require(_mapPath)

	map.x = 0
	map.y = 0
	map.entityType = "other"
	
	map.blockingData = {}
	local width = map.width
	local height = map.height
	
	
	-- returns the mask name string
	function map:getMaskName(index)
	
		-- debug! remove
		--return "mask1111"
		
	--end
	
	--function empty()
		local maskName = "mask"
		local bottom, right, top, left = "0", "0", "0", "0"
		local mapWidth = self.width
		
		-- check bottom
		if map.data[index] == map.data[index + mapWidth] then
			bottom = "1"
		end
		
		-- check right
		if map.data[index] == map.data[index + 1] then
			right = "1"
		end
		
		-- check top
		if map.data[index] == map.data[index - mapWidth] then
			top = "1"
		end
		
		-- check left
		if map.data[index] == map.data[index - 1] then
			left = "1"
		end
		
		maskName = maskName .. bottom .. right .. top .. left
		
		--print("map: mask: " .. maskName)
		return maskName
		
	end
	
	
	--print("Map: create map ... ")
	map.objects = {}
	for h = 0, height - 1 do
		map.blockingData[h] = {}
		
		for w = 0, width - 1 do
			local index = (w + h * width) + 1
			--print("item["..(w+1)..":"..(h+1).."] :"..map.data[index] )
			--self:createMapObject(w, h, map.data[index])
			
			local pos = {x = w * 64, y = h * 64}	-- 64 is the FIXED tile size
			
			--local rndID = math.random(1000,9999)	-- TODO: generate unique IDs
			
			map.objects[table.getn(map.objects) + 1] = objmgr:addObject(terrain:create(pos.x, pos.y))
			
			--
			local _type = map.data[index]
			local objID
			if _type == 1 then
				objID = objmgr:addObject(block:create(pos.x, pos.y))
				
			elseif _type == 2 then
				local maskName = map:getMaskName(index)
				objID = objmgr:addObject(grass:create(pos.x, pos.y, maskName, 1))	-- subtype 1
			
			elseif _type == 3 then
				local maskName = map:getMaskName(index)
				objID = objmgr:addObject(grass:create(pos.x, pos.y, maskName, 2))	-- subtype 2
				
			elseif _type == 4 then
				objID = objmgr:addObject(stone:create(pos.x, pos.y))
				
			elseif _type == 5 then
				local maskName = map:getMaskName(index)
				objID = objmgr:addObject(water:create(pos.x, pos.y, maskName, 1))	-- subtype 1
				
			end
			--
			
			-- build blockingData
			map.blockingData[h][w] = _type
			
			-- fill map data with all objects (tiles)
			map.objects[table.getn(map.objects) + 1] = objID
			
		end
	end
	
	--print("Map: build with size of " .. table.getn(map.objects))
	
	-- initialize map for algorithm
	a_star.setMap({width = map.width, height = map.height, blockingData = map.blockingData})
	
	
	-- return map data at given index
	function map:getMapDataAtPosition(xi, yi)
		-- out of bounds?
		if xi < 0 or xi > self.width - 1 or yi < 0 or yi > self.height - 1 then
			--print("out of bounds!")
			return -1
		end
		
		local index = self:getIndexFlat(xi, yi)
		if map.data[index] ~= nil then
			return map.data[index]
		else
			return -1
		end
	end
	
	
	-- transforn world position into x and y index position
	function map:worldPositionToIndexPosition(_position)
		return math.floor(_position.x / 64), math.floor(_position.y / 64)
	end
	
	
	-- convert 2D index into 1D flat index
	function map:getIndexFlat(xi, yi)
		return (xi + yi * self.width) + 1
	end
	
	--[[
	function map:createMapObject(x, y, type)
		local pos = {X = x * 64, Y = y * 64}	-- 64 is the FIXED tile size
		
		if type == 1 then
			local rndID = math.random(1000,9999)	-- TODO: generate unique IDs
			objmgr:addObject(block:create("block"..rndID, pos.X, pos.Y))
		end
	end
	--]]
	
	-- TODO: the objects are still rendered...
	function map:destroy()
		--local width = map.width
		--local height = map.height
		--print("Map: #objects " .. table.getn(map.objects))
		for i = table.getn(map.objects), 1, -1 do
			--print("Map: destroy object...")
			local obj = objmgr:getObjectByID(map.objects[i])
			if obj then
				print("Map: destroy object...")	-- this does not get called!
				obj:destroy()
			end
		end
		
		-- TODO: invalid!
		--[[
		print("destroy map ... ")
		for h = 0, height - 1 do
			for w = 0, width - 1 do
				local index = self:getIndexFlat(w, h)	--(w + h * width) + 1
				local obj = objmgr:getObjectByID("mapObject" .. index)
				local terrainTile = objmgr:getObjectByID("terrainTile" .. index)
				if obj ~= nil then
					obj:destroy()
				end
				if terrainTile ~= nil then
					terrainTile:destroy()
				end
			end
		end
		--]]
	end
	
	
	function map:findPath(_movementType, start, goal)
		-- pathfinding
		local movementType = _movementType or "walk"
		a_star.setObstacles(tools.getObstaclesByMovementType(movementType))
		--a_star.setMap({width = map.width, height = map.height, blockingData = map.blockingData})
		
		-- start and target x and y in grid coordinates
		path = a_star.run(start, goal)
		if path then
			--[[
			print("path steps: " .. #path)
			for i = #path, 1, -1 do
				local node = path[i]
				print("SOLUTION: [" .. node.x .. ":" .. node.y .. "]")
			end
			--]]
			return path
		else
			print("no path found")
		end
	end
	
	return map
end


return Map