--entity = require "scripts/entity"
mapobject = require "scripts/mapobject"
--imageloader = require "scripts/imageloader"

local Grass = {}

function Grass:create(_x, _y, _maskName, _subtype)
	--print("grass mask: " .. _maskName)
	local grass = mapobject:create({x =_x, y =_y, maskName = _maskName})
	grass.entityType = "grass"
	
	grass.traversableGround = true
	grass.traversableAir = true
	grass.blocking = false
	grass.canSpawnMobs = true
	grass.canSpawnItems = true
	
	grass.images = {}
	grass.animationTick = 0
	grass.currentImageIndex = 1
	
	-- TODO: grass1 is currently static (not animated)
	if _subtype == 1 then
		grass.animationWait = 0.5
		grass.images = imageloader.images["grass1"]
	elseif _subtype == 2 then
		--grass.animationWait = 1.5 + math.random(5) / 10
		grass.animationWait = 1.5
		grass.images = imageloader.images["grass2"]
	end
	
	--grass.image = grass.images[1]

	function grass:tick(dt)
		self.animationTick = self.animationTick + dt
		
		if self.animationTick >= self.animationWait * math.random(15, 25) / 10 then
			self:nextAnimation()
			self.animationTick = 0
		end
	end
	
	
	function grass:nextAnimation()
		self.currentImageIndex = self.currentImageIndex + 1
		if self.images[self.currentImageIndex] == nil then
			self.currentImageIndex = 1
		end
	end
	
	
	if _maskName ~= "mask1111" then	-- mask1111 means this mapobject is encompassed by the same mapobject type on every side!
		function grass:stencilFunction()
			love.graphics.setShader(mask_shader)
			-- cannot use self here for some reason?
			love.graphics.draw(grass.mask, grass.x, grass.y)
			love.graphics.setShader()
		end
	end
	
	
	function grass:draw()
		
		if self.stencilFunction ~= nil then
			--print("grass: set stencil")
			love.graphics.stencil(self.stencilFunction, "replace", 1)
			love.graphics.setStencilTest("greater", 0)
		end
		
		
		-- drawable, x, y, rotation, scale x/y, offset x/y, shearing factor x/y
		love.graphics.draw(self.images[self.currentImageIndex], self.x, self.y, 0, 1, 1, 0, 0)
		
		
		if self.stencilFunction ~= nil then
			love.graphics.setStencilTest()
		end
		
	end
	
	return grass
end


return Grass