local socket = require "socket"
local udp = socket.udp()
-- note: this must be changed on app start (use ip from server/host)
-- TODO: use app:getServerAddress()
local serverAdress = "127.0.0.1"	--"169.172.0.1"
local serverPort = 16118
local updateRate = 0.01
--local checkConnectionsRate = 1
local serverTick = 0	-- reset
local serverTime = 0	-- total time the server is running
local maxPlayers = 2

local Network = {}

-- TODO: buffer messages received?


function Network:init()	-- never gets called atm

	--local network = {}
	
	--gameloop:addLoop(self)
	
end


function Network:tick(dt)

	if self.server then
		self.server.tick(dt)
	end
	
	-- same for client?
	if self.client then
		self.client.tick(dt)
	end
end


function Network:startClient()
	
	self.client = {}
	self.buffer = {}
	
	udp:settimeout(0)
	udp:setpeername(serverAdress, serverPort)
	
	local _ip, _port = udp:getsockname()
	--print("client ip: " .. _ip .. " port: " .. _port)
	self.client.ip = _ip
	self.client.port = _port
	self.client.lastPingTime = 0
	
	app:updatePlayer({ip = _ip, port = _port})
	
	self.client.tick = function(dt)
	
		--print("client: tick")
		
		self.client.listen()
		
		self.client.processBufferData()
		
	end
	
	
	self.client.listen = function()
	
		data, err_msg = udp:receive()
		if data then
			--print("client: Received data: " .. data)
			self.buffer[#self.buffer + 1] = { msg = data, err = err_msg }
			
			
		elseif err_msg ~= "timeout" then
			print("client: Network error: " .. tostring(err_msg))
		end
		
	end
	
	
	self.client.processBufferData = function()
		
		local n = #self.buffer
		
		if n == 0 then
			return
		end
		
		for i = n, 1, - 1 do
		
			local data = self.buffer[i].msg
			
			if data == "check_presence" then
				udp:send("confirm_presence")	-- TODO: queue this? output buffer?
			
			else	-- commands to be further processed...
				
				-- new client got accepted
				if string.find(data, "newclient") ~= nil then
					if string.find(data, self.client.ip) ~= nil
					and string.find(data, self.client.port) ~= nil then
						-- its me!
						print("client: got accepted from server!")
						local human = app:getHumanPlayer()
						udp:send("name:" .. human.displayname)
					end
				
				-- servers sends info about human player at server
				elseif string.find(data, "serverhuman:") ~= nil then
					-- TODO
					local dispname = string.sub(data, 13)
					print("client: server send player info: " .. dispname)
					lobby:addPlayer( {id = 1, displayname = dispname, ip = serverAdress, port = serverPort} )	-- host = ID 1
				
				
				elseif string.find(data, "yourplayerid:") ~= nil then
					local newId = tonumber(string.sub(data, 14))	-- careful: data obtained are strings!
					print("client: server send new id: " .. newId)
					-- update local player id
					app:updatePlayer( {id = newId} )
					local human = app:getHumanPlayer()
					print("current id: ".. newId)
					lobby:addPlayer( {displayname = human.displayname, id = newId, ip = self.client.ip, port = self.client.port} )
				
				-- host started the game; do the same on the client
				elseif string.find(data, "setupgame") ~= nil then
					app:joinGame()
				
				-- server: pong
				elseif data == "pong" then
					local delay = GAMETIME - self.client.lastPingTime
					print("Client: delay is " .. delay)
				
				elseif string.find(data, "mappath:") ~= nil then
					local fileName = string.sub(data, 9)
					lobby:selectMapFile(fileName)
				
				-- entity information (new object created or update)
				elseif string.find(data, "newply:") ~= nil then
					local objInfoString = string.sub(data, 8)
					local objData = tools.string_split(objInfoString, ":")	-- table
					--print("client: server send object info:")
					--print(objData[1])
					--print(objData[2])
					--print(objData[3])
					--print(objData[4])
					-- TODO: validate! + what type of entity to be created/updated? implizit: use objID?
					game:updatePlayerEntity( {x = tonumber(objData[1]), y = tonumber(objData[2]), humanID = tonumber(objData[3]), entityID = objData[4]} )
					-- TODO: different functions to players/mobs/items etc
					--game:updateEntity( {x = tonumber(objData[1]), y = tonumber(objData[2]), humanID = tonumber(objData[3]), entityID = objData[4]} )
				
				-- the server sends his local player entity update (position, rotation)
				elseif string.find(data, "updply:") ~= nil then
					local entityInfoString = string.sub(data, 8)
					local objData = tools.string_split(entityInfoString, ":")
					local entityID = objData[1]
					local x = tonumber(objData[2])
					local y = tonumber(objData[3])
					local rot = tonumber(objData[4])
					--local humanID = self.server.getHumanIdByConnection(msg_or_ip, port_or_nil)
					--if humanID then	-- > 0
						--local entityID = game:getPlayerEntityByHumanID(humanID)
						if entityID then
							local playerEntity = objmgr:getObjectByID(entityID)
							if playerEntity then
								
								-- defined in movingEntity
								playerEntity:setPosition(x, y)
								playerEntity:setRotation(rot)
							end
						end
					--end
					
				end
			
			end
			
			-- processing done
			self.buffer[i] = nil
		end
	end
	
	
	self.client.pingServer = function()
		self.client.lastPingTime = GAMETIME
		udp:send("ping")
	end
	
	
	self.client.sendPlayerKeypress = function(_key)
		udp:send("keypress:" .. _key)
	end
	
	
	-- TODO: move get functions to the code where this function gets called -> here: only relevant info as parameter(s)
	self.client.sendPlayerEntityMovementUpdate = function(_playerEntityTable)
		local entityID = _playerEntityTable.id
		local x, y = _playerEntityTable:getXandY()	-- or getPosition?
		local rot = _playerEntityTable:getRotation()
		udp:send("updply:" .. entityID .. ":" .. x .. ":" .. y .. ":" .. rot)
		--udp:send("updply:" .. x .. ":" .. y .. ":" .. rot)
	end
	
	
	self.client.disconnect = function()
		udp:send("disconnect")
	end
	
	
	udp:send("hello")
	
	gameloop:addLoop(self)
	
	print("network: client started")
	
end

-- TODO: do not need to return lastSeen time?
function Network:getConnections()
	if self.server then
		return self.server.connections
	end
	return {}
end


function Network:startServer()
	
	self.server = {}
	self.buffer = {}
	self.server.active = true
	self.server.connections = {}
	self.server.secondTick = 0
	
	self.server.tick = function(dt)
		--print(tostring(dt))
		serverTime = serverTime + dt
		serverTick = serverTick + dt
		self.server.secondTick = self.server.secondTick + dt
		
		
		if serverTick > updateRate then
			if not self.server.active then return end
			
			--self.server.listen()
			
			--self.server.checkConnectionsPresence()
			
			serverTick = 0
		end
		
		self.server.listen()
		
		self.server.processBufferData()
		
		if self.server.secondTick >= 1 then
			
			-- only once per second
			self.server.checkConnectionsPresence()
			
			self.server.secondTick = 0
		end
		
	end
	
	
	self.server.listen = function()
		
		-- listen to incomming messages
		data, msg_or_ip, port_or_nil = udp:receivefrom()
		if data then
			self.buffer[#self.buffer + 1] = { msg = data, ip = msg_or_ip, port = port_or_nil }
			
		elseif msg_or_ip == "timeout" then
			--print(tostring(serverTime)..": NETWORK TIMEOUT ERROR: " .. tostring(msg_or_ip))
		else
			print(tostring(serverTime)..": NETWORK ERROR: " .. tostring(msg_or_ip))
		end
		
	end
	
	
	self.server.processBufferData = function()
		
		local n = #self.buffer
		
		if n == 0 then
			return
		end
		
		for i = n, 1, - 1 do
		
			local data = self.buffer[i].msg
			local conn = { ip = self.buffer[i].ip, port = self.buffer[i].port }
			--print(tostring(serverTime)..": Received data: " .. data .. " from ip " .. msg_or_ip .." on port " .. port_or_nil)
			
			if data == "hello" then
				-- try to add a new client
				if self.server.newClient(msg_or_ip, port_or_nil) then
					--table.insert(connections, {ip=msg_or_ip, port=port_or_nil})
					--print(tostring(serverTime)..": Received Hello from ip " .. msg_or_ip .." on port " .. port_or_nil)
					
					-- inform all connected clients about the new client (including the new one itself)
					self.server.broadcast("newclient; ip: " .. msg_or_ip .. " port: " .. port_or_nil)
					
				else
					print(tostring(serverTime)..": Denied ip " .. msg_or_ip .." on port " .. port_or_nil .. ": duplicate client or server full!")
				end
			
			-- client is already known ...
			elseif data == "disconnect" then
				self.server.killConnection(conn, "client disconnected")
			
			-- client: ping
			elseif data == "ping" then
				udp:sendto("pong", msg_or_ip, port_or_nil)
			
			elseif data == "confirm_presence" then
				self.server.confirmConnectionPresence(msg_or_ip, port_or_nil)
			
			-- a client sends his local player entity update (position, rotation)
			-- TODO: generalize this for any entity using getObjectByID in the called function
			elseif string.find(data, "updply:") ~= nil then
				local entityInfoString = string.sub(data, 8)
				--print("Server: update player " .. entityInfoString)
				local objData = tools.string_split(entityInfoString, ":")
				local entityID = objData[1]
				local x = tonumber(objData[2])
				local y = tonumber(objData[3])
				local rot = tonumber(objData[4])
				--local humanID = self.server.getHumanIdByConnection(msg_or_ip, port_or_nil)
				--if humanID then	-- > 0
					--local entityID = game:getPlayerEntityByHumanID(humanID)
					if entityID then
						local playerEntity = objmgr:getObjectByID(entityID)
						if playerEntity then
							
							-- defined in movingEntity
							playerEntity:setPosition(x, y)
							playerEntity:setRotation(rot)
						end
					end
				--end
			
			-- a client sends a keypress event info
			elseif string.find(data, "keypress:") ~= nil then
				local key = string.sub(data, 10)
				local humanID = self.server.getHumanIdByConnection(msg_or_ip, port_or_nil)
				if humanID > 0 then
					-- react to this keypress (e.g. try to move the player or shot)
					local entityID = game:getPlayerEntityByHumanID(humanID)
					if entityID then
						local playerEntity = objmgr:getObjectByID(entityID)
						if playerEntity then
							local dt = 0.001	-- TODO: use dt from message?
							if key == "a" then
								-- move!
								playerEntity:accelerate(dt, - playerEntity.maxAcceleration, 0)
							elseif key == "d" then
								playerEntity:accelerate(dt, playerEntity.maxAcceleration, 0)
							elseif key == "w" then
								playerEntity:accelerate(dt, 0, - playerEntity.maxAcceleration)
							elseif key == "s" then
								playerEntity:accelerate(dt, 0, playerEntity.maxAcceleration)
							end
						end
					
					end
					
					-- TODO: answer
					
				end
				
			else
				-- message needs to be further processed
				--print(tostring(serverTime)..": Received Data from ip " .. msg_or_ip .." on port " .. port_or_nil .. ": " .. data)
				
				if string.find(data, "name:") ~= nil then
					--local name = string.gsub(data, "name:", "", 1)
					local name = string.sub(data, 6)
					self.server.setClientName(msg_or_ip, port_or_nil, name)
				end
			end
		
			-- processing done
			self.buffer[i] = nil
		end
		
	end
	
	-- server has to manage player IDs -> server informs every connected player about his (new) ID
	-- server player could always be ID 1? yes!
	self.server.newClient = function(_ip, _port)
	
		local numConnections = #self.server.connections
		if numConnections > (maxPlayers - 1) then
			-- server is already full! do not accept new connections
			print("server: cannot take new client; server is full!")
			return false
		end
		
		if numConnections > 0 then
			for i = 1, numConnections do
				if self.server.connections[i].ip == _ip
				and self.server.connections[i].port == _port then
					-- we already have a client with this address and port, do not add it again!
					return false
				end
			end
		end
		
		-- otherwise (this client is not already known)
		
		-- inform the new client about his player ID
		local newId = self.server.getNewId()
		print("Server: inform client about new ID: " .. newId)
		udp:sendto("yourplayerid:" .. newId, _ip, _port)
		
		-- actually add the client to server connections
		local conn = {id = newId, ip = _ip, port = _port, lastSeen = serverTime}
		table.insert(self.server.connections, conn)
		print("server: added new client with ip " .. _ip .. " and port " .. _port)
		
		-- and inform the new client about the servers human player
		local human = app:getHumanPlayer()
		udp:sendto("serverhuman:" .. human.displayname, _ip, _port)
		
		-- update local lobby
		lobby:addPlayer( {id = newId, displayname = " ", ip = _ip, port = _port} )
		return true
		
	end
	
	
	self.server.getHumanIdByConnection = function(_ip, _port)
		local numConnections = #self.server.connections
		if numConnections > 0 then
			for i = 1, numConnections do
				local conn = self.server.connections[i]
				if conn.ip == _ip
				and conn.port == _port then
					return conn.id	-- = humanID
				end
			end
		end
		
		return 0
	end
	
	
	self.server.getNewId = function()
		local highestId = 1	-- host has already ID 1 -> must be higher!
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.id > highestId then
				highestId = conn.id
			end
		end
		return highestId + 1
	end
	
	
	self.server.broadcast = function(_msg)
		if _msg == nil or _msg == "" then return end
		
		--print("server: broadcast to " .. #self.server.connections .. " connections")
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			udp:sendto(_msg, conn.ip, conn.port)
		end
	end
	
	-- TODO: move get functions to the code where this function gets called -> here: only relevant info as parameter(s)
	self.server.sendPlayerEntityMovementUpdate = function(_playerEntityTable)
		local entityID = _playerEntityTable.id
		local x, y = _playerEntityTable:getXandY()	-- or getPosition?
		local rot = _playerEntityTable:getRotation()
		self.server.broadcast("updply:" .. entityID .. ":" .. x .. ":" .. y .. ":" .. rot)
		--self.server.broadcast("updply:" .. x .. ":" .. y .. ":" .. rot)
	end
	
	
	self.server.checkConnectionsPresence = function()
	
		for i = #self.server.connections, 1, -1 do
			local conn = self.server.connections[i]
			udp:sendto("check_presence", conn.ip, conn.port)
			
			-- server has not received an answer from this client within the last six seconds?
			if serverTime - conn.lastSeen > 6 then
				-- kill this connection!
				self.server.killConnection(conn, "connection timeout; serverTime is " .. serverTime .. " lastSeen is " .. conn.lastSeen)
			end
			
		end
		
	end
	
	-- client answered that it is still there
	self.server.confirmConnectionPresence = function(_ip, _port)
	
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.ip == _ip and conn.port == _port then
				conn.lastSeen = serverTime
			end
		end
	end
	
	
	self.server.killConnection = function(_conn, _reason)
	
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.ip == _conn.ip and conn.port == _conn.port then
				print("server: connection killed: " .. conn.ip .. " (" .. conn.port .. ")")
				print("server: player id is " .. conn.id)
				if _reason ~= nil then
					print("reason: " .. _reason)
				end
				lobby:removePlayerById(conn.id)	-- what if we are not in the lobby?
				
				local uiState = guimgr:getUIstate()
				if uiState == "ingame" or uiState == "ingame_settings" then
					local ply = objmgr:getObjectByID(game:getPlayerByHumanID(conn.id).objID)	-- needs entityID
					if ply then
						ply:destroy()
					end
					game:removePlayerByHumanId(conn.id)
				end
				
				table.remove(self.server.connections, i)
				break
			end
		end
	end
	
	
	self.server.setClientName = function(_ip, _port, _name)
	
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.ip == _ip and conn.port == _port then
				conn.displayname = _name
				-- update local lobby; pass in the connection info
				lobby:setPlayerName(conn)
			end
		end
	end
	
	
	self.server.sendMapInfo = function(fileName)
		self.server.broadcast("mappath:"..fileName)
	end
	
	
	udp:settimeout(0)
	udp:setsockname(serverAdress, serverPort)
	
	local _ip, _port = udp:getsockname()
	app:updatePlayer({ip = _ip, port = _port})
	--lobby:updatePlayer({ip = _ip, port = _port})
	lobby:addPlayer(app:getHumanPlayer())
	
	gameloop:addLoop(self)
	
	print("network: server started")
	
end


function Network:setupGame(mapPath, players)
	-- host started the game
	-- inform all clients so they can start their local game too
	-- params: map path and lobby player table	-- clients should know this already?
	
	self.server.broadcast("setupgame")
	
end

-- TODO: udp:close() ...  does not work the way I suppose it to work (start client, close it and then start server fails because udp is still connected)
function Network:destroy()

	if self.server ~= nil then
		--gameloop:removeLoop(self)
		udp:close()
		self.server = nil
		--print("network: destroyed")
	end
	
	if self.client ~= nil then
		self.client.disconnect()
		udp:close()
		self.client = nil
	end
	gameloop:removeLoop(self)
	print("network: destroyed")
	
end


return Network





--[[
local socket = require "socket"
local udp = socket.udp()
local serverAdress = "127.0.0.1"
local serverPort = 16118
local updateRate = 0.01
--local checkConnectionsRate = 1
local serverTick = 0	-- reset
local serverTime = 0	-- total time the server is running
local maxPlayers = 2

local Network = {}

-- TODO: buffer messages received?


function Network:init()

	--local network = {}
	
	--gameloop:addLoop(self)
	
	self.buffer = {}
	
end


function Network:tick(dt)

	if self.server then
		self.server.tick(dt)
	end
	
	-- same for client?
	if self.client then
		self.client.tick(dt)
	end
end


function Network:startClient()
	
	self.client = {}
	
	udp:settimeout(0)
	udp:setpeername(serverAdress, serverPort)
	
	local _ip, _port = udp:getsockname()
	--print("client ip: " .. _ip .. " port: " .. _port)
	self.client.ip = _ip
	self.client.port = _port
	self.client.lastPingTime = 0
	
	app:updatePlayer({ip = _ip, port = _port})
	
	self.client.tick = function(dt)
	
		--print("client: tick")
		
		self.client.listen()
		
	end
	
	
	self.client.listen = function()
	
		data, err_msg = udp:receive()
		if data then
			--print("client: Received data: " .. data)
			
			if data == "check_presence" then
				udp:send("confirm_presence")
			
			else	-- commands to be further processed...
				
				-- new client got accepted
				if string.find(data, "newclient") ~= nil then
					if string.find(data, self.client.ip) ~= nil
					and string.find(data, self.client.port) ~= nil then
						-- its me!
						print("client: got accepted from server!")
						local human = app:getHumanPlayer()
						udp:send("name:" .. human.displayname)
					end
				
				-- servers sends info about human player at server
				elseif string.find(data, "serverhuman:") ~= nil then
					-- TODO
					local dispname = string.sub(data, 13)
					print("client: server send player info: " .. dispname)
					lobby:addPlayer( {id = 1, displayname = dispname, ip = serverAdress, port = serverPort} )	-- host = ID 1
				
				
				elseif string.find(data, "yourplayerid:") ~= nil then
					local newId = tonumber(string.sub(data, 14))	-- careful: data obtained are strings!
					print("client: server send new id: " .. newId)
					-- update local player id
					app:updatePlayer( {id = newId} )
					local human = app:getHumanPlayer()
					print("current id: ".. newId)
					lobby:addPlayer( {displayname = human.displayname, id = newId, ip = self.client.ip, port = self.client.port} )
				
				-- host started the game; do the same on the client
				elseif string.find(data, "setupgame") ~= nil then
					app:joinGame()
				
				-- server: pong
				elseif data == "pong" then
					local delay = GAMETIME - self.client.lastPingTime
					print("Client: delay is " .. delay)
				
				elseif string.find(data, "mappath:") ~= nil then
					local fileName = string.sub(data, 9)
					lobby:selectMapFile(fileName)
				
				-- entity information (new object created or update)
				elseif string.find(data, "newply:") ~= nil then
					local objInfoString = string.sub(data, 8)
					local objData = tools.string_split(objInfoString, ":")	-- table
					--print("client: server send object info:")
					--print(objData[1])
					--print(objData[2])
					--print(objData[3])
					--print(objData[4])
					-- TODO: validate! + what type of entity to be created/updated? implizit: use objID?
					game:updatePlayerEntity( {x = tonumber(objData[1]), y = tonumber(objData[2]), humanID = tonumber(objData[3]), entityID = objData[4]} )
					-- TODO: different functions to players/mobs/items etc
					--game:updateEntity( {x = tonumber(objData[1]), y = tonumber(objData[2]), humanID = tonumber(objData[3]), entityID = objData[4]} )
				
				-- the server sends his local player entity update (position, rotation)
				elseif string.find(data, "updply:") ~= nil then
					local entityInfoString = string.sub(data, 8)
					local objData = tools.string_split(entityInfoString, ":")
					local entityID = objData[1]
					local x = tonumber(objData[2])
					local y = tonumber(objData[3])
					local rot = tonumber(objData[4])
					--local humanID = self.server.getHumanIdByConnection(msg_or_ip, port_or_nil)
					--if humanID then	-- > 0
						--local entityID = game:getPlayerEntityByHumanID(humanID)
						if entityID then
							local playerEntity = objmgr:getObjectByID(entityID)
							if playerEntity then
								
								-- defined in movingEntity
								playerEntity:setPosition(x, y)
								playerEntity:setRotation(rot)
							end
						end
					--end
					
				end
			
			end
			
		elseif err_msg ~= "timeout" then
			print("client: Network error: " .. tostring(err_msg))
		end
		
	end
	
	
	self.client.pingServer = function()
		self.client.lastPingTime = GAMETIME
		udp:send("ping")
	end
	
	
	self.client.sendPlayerKeypress = function(_key)
		udp:send("keypress:" .. _key)
	end
	
	
	-- TODO: move get functions to the code where this function gets called -> here: only relevant info as parameter(s)
	self.client.sendPlayerEntityMovementUpdate = function(_playerEntityTable)
		local entityID = _playerEntityTable.id
		local x, y = _playerEntityTable:getXandY()	-- or getPosition?
		local rot = _playerEntityTable:getRotation()
		udp:send("updply:" .. entityID .. ":" .. x .. ":" .. y .. ":" .. rot)
		--udp:send("updply:" .. x .. ":" .. y .. ":" .. rot)
	end
	
	
	self.client.disconnect = function()
		udp:send("disconnect")
	end
	
	
	udp:send("hello")
	
	gameloop:addLoop(self)
	
	print("network: client started")
	
end

-- TODO: do not need to return lastSeen time?
function Network:getConnections()
	if self.server then
		return self.server.connections
	end
	return {}
end


function Network:startServer()
	
	self.server = {}
	self.server.active = true
	self.server.connections = {}
	self.server.secondTick = 0
	
	self.server.tick = function(dt)
		--print(tostring(dt))
		serverTime = serverTime + dt
		serverTick = serverTick + dt
		self.server.secondTick = self.server.secondTick + dt
		
		if serverTick > updateRate then
			if not self.server.active then return end
			
			self.server.listen()
			
			--self.server.checkConnectionsPresence()
			
			serverTick = 0
		end
		
		if self.server.secondTick >= 1 then
			
			-- only once per second
			self.server.checkConnectionsPresence()
			
			self.server.secondTick = 0
		end
		
	end
	
	
	self.server.listen = function()
		
		-- listen to incomming messages
		data, msg_or_ip, port_or_nil = udp:receivefrom()
		if data then
			local conn = {ip = msg_or_ip, port = port_or_nil}
			--print(tostring(serverTime)..": Received data: " .. data .. " from ip " .. msg_or_ip .." on port " .. port_or_nil)
			
			if data == "hello" then
				-- try to add a new client
				if self.server.newClient(msg_or_ip, port_or_nil) then
					--table.insert(connections, {ip=msg_or_ip, port=port_or_nil})
					--print(tostring(serverTime)..": Received Hello from ip " .. msg_or_ip .." on port " .. port_or_nil)
					
					-- inform all connected clients about the new client (including the new one itself)
					self.server.broadcast("newclient; ip: " .. msg_or_ip .. " port: " .. port_or_nil)
					
				else
					print(tostring(serverTime)..": Denied ip " .. msg_or_ip .." on port " .. port_or_nil .. ": duplicate client or server full!")
				end
			
			-- client is already known ...
			elseif data == "disconnect" then
				self.server.killConnection(conn, "client disconnected")
			
			-- client: ping
			elseif data == "ping" then
				udp:sendto("pong", msg_or_ip, port_or_nil)
			
			elseif data == "confirm_presence" then
				self.server.confirmConnectionPresence(msg_or_ip, port_or_nil)
			
			-- a client sends his local player entity update (position, rotation)
			-- TODO: generalize this for any entity using getObjectByID in the called function
			elseif string.find(data, "updply:") ~= nil then
				local entityInfoString = string.sub(data, 8)
				--print("Server: update player " .. entityInfoString)
				local objData = tools.string_split(entityInfoString, ":")
				local entityID = objData[1]
				local x = tonumber(objData[2])
				local y = tonumber(objData[3])
				local rot = tonumber(objData[4])
				--local humanID = self.server.getHumanIdByConnection(msg_or_ip, port_or_nil)
				--if humanID then	-- > 0
					--local entityID = game:getPlayerEntityByHumanID(humanID)
					if entityID then
						local playerEntity = objmgr:getObjectByID(entityID)
						if playerEntity then
							
							-- defined in movingEntity
							playerEntity:setPosition(x, y)
							playerEntity:setRotation(rot)
						end
					end
				--end
			
			-- a client sends a keypress event info
			elseif string.find(data, "keypress:") ~= nil then
				local key = string.sub(data, 10)
				local humanID = self.server.getHumanIdByConnection(msg_or_ip, port_or_nil)
				if humanID > 0 then
					-- react to this keypress (e.g. try to move the player or shot)
					local entityID = game:getPlayerEntityByHumanID(humanID)
					if entityID then
						local playerEntity = objmgr:getObjectByID(entityID)
						if playerEntity then
							local dt = 0.001	-- TODO: use dt from message?
							if key == "a" then
								-- move!
								playerEntity:accelerate(dt, - playerEntity.maxAcceleration, 0)
							elseif key == "d" then
								playerEntity:accelerate(dt, playerEntity.maxAcceleration, 0)
							elseif key == "w" then
								playerEntity:accelerate(dt, 0, - playerEntity.maxAcceleration)
							elseif key == "s" then
								playerEntity:accelerate(dt, 0, playerEntity.maxAcceleration)
							end
						end
					
					end
					
					-- TODO: answer
					
				end
				
			else
				-- message needs to be further processed
				--print(tostring(serverTime)..": Received Data from ip " .. msg_or_ip .." on port " .. port_or_nil .. ": " .. data)
				
				if string.find(data, "name:") ~= nil then
					--local name = string.gsub(data, "name:", "", 1)
					local name = string.sub(data, 6)
					self.server.setClientName(msg_or_ip, port_or_nil, name)
				end
			end
			
		elseif msg_or_ip == "timeout" then
			--print(tostring(serverTime)..": NETWORK TIMEOUT ERROR: " .. tostring(msg_or_ip))
		else
			print(tostring(serverTime)..": NETWORK ERROR: " .. tostring(msg_or_ip))
		end
		
	end
	
	-- server has to manage player IDs -> server informs every connected player about his (new) ID
	-- server player could always be ID 1? yes!
	self.server.newClient = function(_ip, _port)
	
		local numConnections = #self.server.connections
		if numConnections > (maxPlayers - 1) then
			-- server is already full! do not accept new connections
			print("server: cannot take new client; server is full!")
			return false
		end
		
		if numConnections > 0 then
			for i = 1, numConnections do
				if self.server.connections[i].ip == _ip
				and self.server.connections[i].port == _port then
					-- we already have a client with this address and port, do not add it again!
					return false
				end
			end
		end
		
		-- otherwise (this client is not already known)
		
		-- inform the new client about his player ID
		local newId = self.server.getNewId()
		print("Server: inform client about new ID: " .. newId)
		udp:sendto("yourplayerid:" .. newId, _ip, _port)
		
		-- actually add the client to server connections
		local conn = {id = newId, ip = _ip, port = _port, lastSeen = serverTime}
		table.insert(self.server.connections, conn)
		print("server: added new client with ip " .. _ip .. " and port " .. _port)
		
		-- and inform the new client about the servers human player
		local human = app:getHumanPlayer()
		udp:sendto("serverhuman:" .. human.displayname, _ip, _port)
		
		-- update local lobby
		lobby:addPlayer( {id = newId, displayname = " ", ip = _ip, port = _port} )
		return true
		
	end
	
	
	self.server.getHumanIdByConnection = function(_ip, _port)
		local numConnections = #self.server.connections
		if numConnections > 0 then
			for i = 1, numConnections do
				local conn = self.server.connections[i]
				if conn.ip == _ip
				and conn.port == _port then
					return conn.id	-- = humanID
				end
			end
		end
		
		return 0
	end
	
	
	self.server.getNewId = function()
		local highestId = 1	-- host has already ID 1 -> must be higher!
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.id > highestId then
				highestId = conn.id
			end
		end
		return highestId + 1
	end
	
	
	self.server.broadcast = function(_msg)
		if _msg == nil or _msg == "" then return end
		
		--print("server: broadcast to " .. #self.server.connections .. " connections")
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			udp:sendto(_msg, conn.ip, conn.port)
		end
	end
	
	-- TODO: move get functions to the code where this function gets called -> here: only relevant info as parameter(s)
	self.server.sendPlayerEntityMovementUpdate = function(_playerEntityTable)
		local entityID = _playerEntityTable.id
		local x, y = _playerEntityTable:getXandY()	-- or getPosition?
		local rot = _playerEntityTable:getRotation()
		self.server.broadcast("updply:" .. entityID .. ":" .. x .. ":" .. y .. ":" .. rot)
		--self.server.broadcast("updply:" .. x .. ":" .. y .. ":" .. rot)
	end
	
	
	self.server.checkConnectionsPresence = function()
	
		for i = #self.server.connections, 1, -1 do
			local conn = self.server.connections[i]
			udp:sendto("check_presence", conn.ip, conn.port)
			
			-- server has not received an answer from this client within the last six seconds?
			if serverTime - conn.lastSeen > 6 then
				-- kill this connection!
				self.server.killConnection(conn, "connection timeout; serverTime is " .. serverTime .. " lastSeen is " .. conn.lastSeen)
			end
			
		end
		
	end
	
	-- client answered that it is still there
	self.server.confirmConnectionPresence = function(_ip, _port)
	
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.ip == _ip and conn.port == _port then
				conn.lastSeen = serverTime
			end
		end
	end
	
	
	self.server.killConnection = function(_conn, _reason)
	
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.ip == _conn.ip and conn.port == _conn.port then
				print("server: connection killed: " .. conn.ip .. " (" .. conn.port .. ")")
				print("server: player id is " .. conn.id)
				if _reason ~= nil then
					print("reason: " .. _reason)
				end
				lobby:removePlayerById(conn.id)	-- what if we are not in the lobby?
				
				local uiState = guimgr:getUIstate()
				if uiState == "ingame" or uiState == "ingame_settings" then
					local ply = objmgr:getObjectByID(game:getPlayerByHumanID(conn.id).objID)	-- needs entityID
					if ply then
						ply:destroy()
					end
					game:removePlayerByHumanId(conn.id)
				end
				
				table.remove(self.server.connections, i)
				break
			end
		end
	end
	
	
	self.server.setClientName = function(_ip, _port, _name)
	
		for i = 1, #self.server.connections do
			local conn = self.server.connections[i]
			if conn.ip == _ip and conn.port == _port then
				conn.displayname = _name
				-- update local lobby; pass in the connection info
				lobby:setPlayerName(conn)
			end
		end
	end
	
	
	self.server.sendMapInfo = function(fileName)
		self.server.broadcast("mappath:"..fileName)
	end
	
	
	udp:settimeout(0)
	udp:setsockname(serverAdress, serverPort)
	
	local _ip, _port = udp:getsockname()
	app:updatePlayer({ip = _ip, port = _port})
	--lobby:updatePlayer({ip = _ip, port = _port})
	lobby:addPlayer(app:getHumanPlayer())
	
	gameloop:addLoop(self)
	
	print("network: server started")
	
end


function Network:setupGame(mapPath, players)
	-- host started the game
	-- inform all clients so they can start their local game too
	-- params: map path and lobby player table	-- clients should know this already?
	
	self.server.broadcast("setupgame")
	
end

-- TODO: udp:close() ...  does not work the way I suppose it to work (start client, close it and then start server fails because udp is still connected)
function Network:destroy()

	if self.server ~= nil then
		--gameloop:removeLoop(self)
		udp:close()
		self.server = nil
		--print("network: destroyed")
	end
	
	if self.client ~= nil then
		self.client.disconnect()
		udp:close()
		self.client = nil
	end
	gameloop:removeLoop(self)
	print("network: destroyed")
	
end


return Network

--]]