-- this is the object manager

local ObjMgr = {}


function ObjMgr:load()
	self.highestID = 0
	self.objects = {}
	
	-- create a subtable for every entityType
	self.objects["player"] = {}
	self.objects["bat"] = {}
	self.objects["spider"] = {}
	self.objects["zombie"] = {}
	self.objects["bullet"] = {}
	self.objects["grass"] = {}
	self.objects["block"] = {}
	self.objects["stone"] = {}
	self.objects["water"] = {}
	self.objects["terrain"] = {}
	self.objects["other"] = {}
	
	--gameloop:addLoop(self)
end


function ObjMgr:tick(dt)


end

-- multiplayer: force to use a specific ID (this should be unique!?)
--> returns the passed in _forcedID
function ObjMgr:addObject(obj, _forcedID)
	if obj.entityType ~= nil then
		--print("obj type: " .. obj.entityType)
	else
		--error(obj)
		--print("OBJEKT HAS NO ENTITYTYPE")
	end
	if obj.entityCategory ~= nil then
		--print("obj category: " .. obj.entityCategory)
	end
	
	--print("objmgr: new pos: " .. (obj.x or "?") .. ":" ..(obj.y or "?"))
	
	local entType = obj.entityType or "other"
	local id
	if _forcedID then
		id = _forcedID
	else
		id = ObjMgr:getNewID(entType)
	end
	--print("objmgr: new id is " .. id)
	if obj.load == nil then
		print("note: object \""..id.."\" has no load function!")
	else
		obj:load(id)
	end
	table.insert(self.objects[entType], obj)
	
	return id
end	


function ObjMgr:getNewID(_entityType)
	if self.objects[_entityType] == nil then
		print("objmgr error: unknonw entity type: ")
		print(_entityType)
		return
	end
	
	-- TODO: note that this value is unlimitted at the moment! theoretically this could result in ??? errors when this number gets too large
	self.highestID = self.highestID + 1
	return _entityType .. "_" .. self.highestID	-- the "_" acts as delimiter between entityType and the "actual" ID (number)
	
end


function ObjMgr:getEntityTypeByID(id)

	-- id looks like this: "entityType_number"
	local t = tools.string_split(id, "_")
	return t[1]
	
end


function ObjMgr:getObjectByID(id)

	local _entityType = ObjMgr:getEntityTypeByID(id)
	
	if self.objects[_entityType] == nil then
		print("objmgr error: unknonw entity type: ")
		print(_entityType)
		return
	end
	
	for i = 1, #self.objects[_entityType] do
		local _obj = self.objects[_entityType][i]
		if _obj.id == id then
			return _obj
		end
	end

	return nil
end


-- destroy object by ID
function ObjMgr:destroyObject(id)
	local _entityType = ObjMgr:getEntityTypeByID(id)
	
	if self.objects[_entityType] == nil then
		print("objmgr error: unknonw entity type: ")
		print(_entityType)
		return false
	end
	
	for i = 1, #self.objects[_entityType] do
		local _obj = self.objects[_entityType][i]
		if _obj.id == id then
			table.remove(self.objects[_entityType], i)
			return true
		end
	end
	return false
end


--[[
-- destroy by ID or pass the table
function ObjMgr:destroyObject(obj)
	for i = #self.objects,1,-1 do
		local object = self.objects[i]
		if (type(obj) == "table") and (obj == object) then
			--print("destroying object by table ...")
			table.remove(self.objects, i)
			obj = nil
			break
		elseif ((type(obj) == "string") or (type(obj) == "number")) and (obj == object.id) then
			--print("destroying object by id ...")
			table.remove(self.objects, i)
			obj = nil
			break
		end
	end
end
--]]


function ObjMgr:getobjectcount()
	return #self.objects
end


return ObjMgr
