local Renderer = {}

local numLayers = 8

LAYERTERRAIN = 1
LAYERMAPOBJECTS = 2
LAYERITEMS = 3
LAYERMOBS = 4
LAYERPLAYERS = 5
LAYEREFFECTS = 6
LAYEROVERHEAD = 7
LAYERGUI = 8

--[[
layer 1: terrain
layer 2: background / map objects (walls etc)
layer 3: items
layer 4: enemies / mobs
layer 5: players
layer 6: bullets
layer 7: overhead (healthbars)
layer 8: GUI texts etc
--]]

function Renderer:load()
	self.renderers = {}
	for i = 1, numLayers do
		self.renderers[i] = {}
	end
end

function Renderer:addRenderer(obj, layer)
	local l = layer or 4
	if obj.visible == nil then obj.visible = true end
	table.insert(self.renderers[l], obj)
end

function Renderer:removeRenderer(obj)
	for layer = numLayers, 1, -1 do
		for o = #self.renderers[layer], 1, -1 do
			if self.renderers[layer][o] == obj then
				table.remove(self.renderers[layer], o)
				--print("remove renderer!")
				return
			end
		end
	end
end

-- draw layer 1, draw layer 2, draw layer 3 ...
function Renderer:draw()
	cam:set()
	
	--love.graphics.push()
	--love.graphics.setShader(test_shader)
	--mask_shader:send("isactive", true)
	
	for layer = 1, numLayers - 1 do
		for i = 1, #self.renderers[layer] do
			local obj = self.renderers[layer][i]
			if obj.visible then
				--love.graphics.push()
				obj:draw(dt)
				--love.graphics.pop()
			end
		end
	end
	
	cam:unset()
	
	--love.graphics.pop()
	--love.graphics.setShader(test_shader)
	--mask_shader:send("isactive", false)

	-- draw UI
	for i = 1, #self.renderers[numLayers] do
		local obj = self.renderers[numLayers][i]
		if obj.visible then
			obj:draw(dt)
		end
	end
end


return Renderer