map = require "scripts/map"

local World = {}

function World:create(_mapPath)

	local world = {}
	world.x = 0
	world.y = 0
	world.entityType = "other"
	
	--world.size = {height = 64, width = 64}
	world.options = {	-- add some features later (environment, general settings, spawn chances, drop rates etc?)
		mobLimit = 0,				-- temporary limit (mob count at once)
		totalMobLimit = 20,			-- when this amount of mobs have been spawned, do not spawn any more, TODO: not implemented yet
		playerSafeDistance = 400,	-- do not spawn mobs too close to players
		maxSpawnAttempts = 10,		-- if the randomly selected spawn position cannot be used right now, try again until this limit is reached
	}
	world.mapPath = _mapPath
	world.players = {}		-- a human player ID (key) gets the player entityID (value)
	world.playerCount = 0	-- note that this is and indexed table, so #table wont work to get its size!
	
	-- they only hold their entity IDs!
	world.mobs = {}
	world.items = {}
	world.projectiles = {}
	
	world.mobSpawnBehaviour = {}
	world.mobSpawnBehaviour["spider"]		= {chance = 0.0, groupMin = 1, groupMax = 3}
	world.mobSpawnBehaviour["bat"]			= {chance = 0.0, groupMin = 1, groupMax = 4}
	world.mobSpawnBehaviour["zombie"]		= {chance = 0.2, groupMin = 1, groupMax = 2}
	
	--world.lastMobSpawnTime = 0
	world.secondTick = 0
	world.deciSecondTick = 0
	
	print("world: init world with map: ".._mapPath)
	world.map = map:create(_mapPath)
	
	world.playerSpawns = {}
	for i = 1, #world.map.playerSpawns do
		world.playerSpawns[i] = {}
		world.playerSpawns[i].x = world.map.playerSpawns[i].x
		world.playerSpawns[i].y = world.map.playerSpawns[i].y
		world.playerSpawns[i].pID = 0
	end
	
	
	function world:load(_id)
		self.id = _id
		gameloop:addLoop(self)
		
		--print("init world with map: ".._mapPath)
		--world.map = map:create(_mapPath)
	end

	
	function world:tick(dt)
		
		self:controlCreatures(dt)
		
		self:controlItemSpawn(dt)
		
		self:controlProjectiles(dt)
		
		
		-- perform some task once per second only
		--self:triggerEverySecond(dt)
		
		-- 10x per second
		self:triggerEveryDeciSecond(dt)
		
	end
	
	
	function world:triggerEverySecond(dt)
	
		self.secondTick = self.secondTick + dt
		
		-- do not do this more than once per second
		if self.secondTick >= 1 then
		
			-- execute everything that only needs to run once per second
			--self:controlMobSpawn(dt)
			--self:controlMobMovement(dt)
			
			self.secondTick = 0
		end
		
	end
	
	
	function world:triggerEveryDeciSecond(dt)
	
		self.deciSecondTick = self.deciSecondTick + dt
		
		-- do not do this more than once per decisecond
		if self.deciSecondTick >= 0.1 then
		
			-- execute everything that only needs to run 10 times per second
			self:controlMobSpawn(dt)
			self:controlMobMovement(dt)
			
			self.deciSecondTick = 0
		end
		
	end
	
	
	-- TODO: the game needs the human player info, world just needs the player entity
	-- now: (human player ID, entityID)
	function world:addPlayer(humanID, id)
		--print(_player.id)
		if (humanID == nil) or (id == nil) then
			print("error in world: new player has no ID")
			return
		end
		
		if self.players[humanID] ~= nil then
			print("error in world: new player with ID " .. humanID .. " is already existing in this world!")
			return
		end
		
		self.players[humanID] = id
		self.playerCount = self.playerCount + 1
		
	end
	
	
	--function world:getPlayerEntityByHumanID(_humanID)
		
	--end
	
	
	-- TODO: all places occupied? nil return = error!
	function world:getPlayerSpawnPosition(_playerID)
		--print("World: search spawn for player " .. _playerID)
		local spawns = self.playerSpawns
		--[[
		local i = math.random(1, #spawns)
		return spawns[i].X, spawns[i].Y
		--]]
		--
		for i = 1, #spawns do
			-- random? TODO
			
			if spawns[i].playerID == nil then
				spawns[i].playerID = _playerID
				return spawns[i].x, spawns[i].y
			end
		end
		--
	end
	
	
	function world:getMapObjectTypeAtPosition(_position)
		local xi, yi = math.floor(_position.x / 64), math.floor(_position.y / 64)
		
		if xi > -1 and yi > -1 and xi <= self.map.width and yi <= self.map.height then
			
			--local xi, yi = math.floor(_position.x / 64), math.floor(_position.y / 64)
			--local index = self.map:getIndexFlat(xi, yi) --(xi + yi * self.map.width) + 1
			return self.map:getMapDataAtPosition(xi, yi)
		else
			return -1
		end
	end
	
	
	function world:getEntityEnvironment(_position)
		local xi, yi = math.floor(_position.x / 64), math.floor(_position.y / 64)
		local env = {}
		env[1] = {x = (xi - 1) * 64, y = (yi - 1) * 64, type = self.map:getMapDataAtPosition(xi - 1, yi - 1), index = self.map:getIndexFlat(xi - 1, yi - 1)}
		env[2] = {x = (xi    ) * 64, y = (yi - 1) * 64, type = self.map:getMapDataAtPosition(xi    , yi - 1), index = self.map:getIndexFlat(xi    , yi - 1)}
		env[3] = {x = (xi + 1) * 64, y = (yi - 1) * 64, type = self.map:getMapDataAtPosition(xi + 1, yi - 1), index = self.map:getIndexFlat(xi + 1, yi - 1)}
		env[4] = {x = (xi - 1) * 64, y = (yi    ) * 64, type = self.map:getMapDataAtPosition(xi - 1, yi    ), index = self.map:getIndexFlat(xi - 1, yi    )}
		env[5] = {x = (xi    ) * 64, y = (yi    ) * 64, type = self.map:getMapDataAtPosition(xi    , yi    ), index = self.map:getIndexFlat(xi    , yi    )}
		env[6] = {x = (xi + 1) * 64, y = (yi    ) * 64, type = self.map:getMapDataAtPosition(xi + 1, yi    ), index = self.map:getIndexFlat(xi + 1, yi    )}
		env[7] = {x = (xi - 1) * 64, y = (yi + 1) * 64, type = self.map:getMapDataAtPosition(xi - 1, yi + 1), index = self.map:getIndexFlat(xi - 1, yi + 1)}
		env[8] = {x = (xi    ) * 64, y = (yi + 1) * 64, type = self.map:getMapDataAtPosition(xi    , yi + 1), index = self.map:getIndexFlat(xi    , yi + 1)}
		env[9] = {x = (xi + 1) * 64, y = (yi + 1) * 64, type = self.map:getMapDataAtPosition(xi + 1, yi + 1), index = self.map:getIndexFlat(xi + 1, yi + 1)}
		
		return env
	end
	
	
	-- for all creatures
	function world:controlCreatures(dt)
		
		-- check if a player or mob is about to collide with a blocking / not traversable map object
		--local numberOfPlayers = self.playerCount
		--print(numberOfPlayers)
		--for i = 1, numberOfPlayers do
		for _p, _playerID in pairs(self.players) do
			--print("world player: " .. _playerID)
			local _player = objmgr:getObjectByID(_playerID)
			--print("world player id: " .. _player.id)
			--local posX, posY = _player:getPosition()
			--print("...pos: " .. posX ..":"..posY)
			
			-- somehow the world must find out if a player is still alive
			--local alive = _player:isAlive()
			if _player == nil then
				print("world: player went invalid (remove)")
				self.players[_p] = nil
				
			else
			
				world:checkEnvironmentForCreature(_player)
			
				-- fetch bullets to be created, fired by this player
				local bulletTable = _player.bulletsToCreate
				local numBullets = #bulletTable
				if numBullets > 0 then
					for i, bullet in pairs(bulletTable) do
						self:createProjectile(bullet)	-- TODO: who to hurt? whom to give credit?
						table.remove(bulletTable, i)
					end
				end
				
			end
			
			
			
		end
		
		-- check for mobs
		for _, _mobID in pairs(self.mobs) do
			local _mob = objmgr:getObjectByID(_mobID)
			world:checkEnvironmentForCreature(_mob)
			
			-- fetch closest player for every mob
			local mobX, mobY = _mob:getPosition()
			local closestPlayer = self:getClosestPlayerIDByPosition(mobX, mobY, _mob.followPlayerDistance)
			if closestPlayer then
				_mob:setAttackTarget(closestPlayer)
			end
			
			-- fetch bullets to be created, fired by this mob
			-- ...
			
		end
		
	end
	
	
	function world:checkEnvironmentForCreature(_creature)
		--print(k, v)
		--
		local posX, posY = _creature:getPosition()
		local pos = {x = math.floor(posX), y = math.floor(posY)}
		--print("testing for position: " .. pos.x .. " : " .. pos.y)
		--local mapObject = world:getMapObjectTypeAtPosition(pos)-- or 0
		--print("map object under player is: " .. mapObject)
		
		-- next: get the surrounding map object TYPES (neighbours) -> simply pass it to the creature, do not check for collisions HERE
		local entityEnvironment = world:getEntityEnvironment(pos)
		--[[
		print(entityEnvironment[1].type .. " " .. entityEnvironment[2].type .. " " .. entityEnvironment[3].type)
		print(entityEnvironment[4].type .. " " .. entityEnvironment[5].type .. " " .. entityEnvironment[6].type)
		print(entityEnvironment[7].type .. " " .. entityEnvironment[8].type .. " " .. entityEnvironment[9].type)
		print(" ")
		--]]
		
		--local colliding = false
		_creature.collisionPoint = nil
		
		local playerCollisionEnvironment = {}
		
		for k, envObject in pairs(entityEnvironment) do
			--if tools.isMapObjectTypeBlocking(envObject.type) then
			if tools.isMapObjectTypeBlockingByMovementType(envObject.type, _creature.movementType) then
				local rect = {x = envObject.x, y = envObject.y, height = 64, width = 64}
				--print("world insert rect " .. rect.x .. ":" .. rect.y)
				table.insert(playerCollisionEnvironment, rect)
				--[[
				local obj = objmgr:getObjectByID("mapObject"..envObject.index)
				if obj ~= nil then
					local rect = {x = obj.x, y = obj.y, height = 64, width = 64}
					table.insert(playerCollisionEnvironment, rect)
				end
				--]]
				
				--[[
				-- do the collision check
				--local px, py = _player:getXandY()
				local px, py = _player:precalculatePosition(dt)
				local circle = {radius = _player.size / 2, x = px, y = py}
				local rect = {x = obj.x, y = obj.y, height = 64, width = 64}
				--if tools.areCircleAndRectColliding(circle, rect) then
				local point = tools.getCollisionPoint(circle, rect)
				if point ~= nil then
					--print("world: collision with index " .. k)
					--_player:hurt(0.1)
					_player:collideAtPoint(point, dt)
					--colliding = true
					
				end
				--]]
			else
				--print("non-blocking " .. k)
			end
			
			--_player.colliding = colliding
		end
		
		-- finally, give the blocking environment information to the player
		_creature.environment = playerCollisionEnvironment
		
		--local px, py = _player:getXandY()
		--print("player pos: " .. pos.x .. ":" .. pos.y .. "corner: " .. px .. ":" .. py)
		
	end
	
	
	function world:createProjectile(_bullet)
	
		--print("world: create bullet:")
		
		-- actually create the bullet
		local newBullet = bullet:create(_bullet.x, _bullet.y, _bullet.movingVector, _bullet.sourceID, _bullet.size, _bullet.speed, _bullet.damage, _bullet.rotation)
		local id = objmgr:addObject(newBullet)
		self:addProjectile(id)
		
	end
	
	
	function world:addProjectile(projectile)
		table.insert(self.projectiles, projectile)
	end
	
	
	function world:controlProjectiles(dt)
		for i, projectileID in pairs(self.projectiles) do
			local projectile = objmgr:getObjectByID(projectileID)
			-- check for collisions here: send info to projectile so that it can fire its collision function
			local posX, posY = projectile:getPosition()
			local pos = {x = math.floor(posX), y = math.floor(posY)}
			
			-- projectile colliding with a wall? destroy it
			local entityEnvironment = world:getEntityEnvironment(pos)
			for k, envObject in pairs(entityEnvironment) do
				if tools.isMapObjectTypeBlockingByMovementType(envObject.type, projectile.movementType) then
					local circle = {x = posX, y = posY, radius = projectile.size/2}
					local rect = {x = envObject.x, y = envObject.y, height = 64, width = 64}
					if tools.areCircleAndRectColliding(circle, rect) then
						--print("world: projectile collided with wall")
						projectile:destroy()
						table.remove(self.projectiles, i)
						break
					end
					--[[
					local obj = objmgr:getObjectByID("mapObject"..envObject.index)
					if obj ~= nil then
						local circle = {x = posX, y = posY, radius = projectile.size/2}
						local rect = {x = obj.x, y = obj.y, height = 64, width = 64}
						if tools.areCircleAndRectColliding(circle, rect) then
							--print("world: projectile collided with wall")
							projectile:destroy()
							table.remove(self.projectiles, i)
							break
						end
					end
					--]]
				end
			end
			
			-- projectile hitting a creature? TODO: what creatures trigger a hit?
			for j, _mobID in pairs(self.mobs) do
				local mob = objmgr:getObjectByID(_mobID)
				local mobX, mobY = mob:getPosition()
				local circle_1 = {x = posX, y = posY, radius = projectile.size/2}
				local circle_2 = {x = mobX, y = mobY, radius = mob.size/2}
				if tools.areCirclesColliding(circle_1, circle_2) then
					---print("world: projectile collided with mob")
					-- hurt the mob
					if mob:isDamageLethal(projectile.damage) then
						table.remove(self.mobs, j)
						--print("world: remove mob " .. mob.id)
					end
					mob:hurt(projectile.damage, projectile.sourceID)
					
					-- destroy the projectile
					projectile:destroy()
					table.remove(self.projectiles, i)
					break
				end
			end
		end
	end
	
	
	function world:addMob(mobID)	-- takes the ID only!
		table.insert(self.mobs, mobID)
		--print("world mob count: " .. #self.mobs)
	end
	
	
	function world:getMobCount()
		return #self.mobs
	end
	
	
	function world:controlMobSpawn(dt)
	
		--[[
		-- do not do this more than once per second
		if GAMETIME - self.lastMobSpawnTime < 1 then
			return
		end
		--]]
		
		-- do not spawn any mobs if limit is reached
		local availableMobs = self.options.mobLimit - self:getMobCount()
		if availableMobs < 1 then
			return
		end
		
		-- do the mob spawning ...
		for mobType, spawnData in pairs(self.mobSpawnBehaviour) do
		
			local chance = math.random()
			if chance > (1 - spawnData.chance) then
				
				--local spawnPosX, spawnPosY = 2, 2	-- TODO
				local spawnAmount = math.min(math.random(spawnData.groupMin, spawnData.groupMax), availableMobs)
				
				for i = 1, spawnAmount do
					self:spawnMob(mobType)	-- returns true on success
				end
				
			end
		
		end
	end
	
	
	function world:getSpawnChance(_mobType)
		return self.mobSpawnBehaviour[_mobType].chance
	end
	
	-- set new spawn chance for a given mob type, returns new value on success or -1 on failure
	function world:setSpawnChance(_mobType, _chance)
		if self.mobSpawnBehaviour[_mobType] == nil then
			return -1
		else
			local newChance = tools.clamp(_chance, 0, 1)
			self.mobSpawnBehaviour[_mobType].chance = newChance
			return newChance
		end
	end
	
	
	function world:getMobSpawnPosition(_movementType, _attempt)
		local rnd = math.random(1, #self.map.mobSpawns[_movementType])
		local spawnPosRaw = self.map.mobSpawns[_movementType][rnd]	-- this is in grid coordinates!
		local spawnPos = {}
		spawnPos.x = (spawnPosRaw.x - 1) * 64
		spawnPos.y = (spawnPosRaw.y - 1) * 64
		--print("world initial spawn pos = " .. spawnPos.x .. ":" .. spawnPos.y)
		local satisfySafeSpawn = true
		for _, _playerID in pairs(self.players) do
			local _player = objmgr:getObjectByID(_playerID)
			local playerX, playerY = _player:getPosition()
			local playerPos = {x = playerX, y = playerY}
			local distanceSquared = tools.getDistanceSquared(spawnPos, playerPos)
			if distanceSquared < (self.options.playerSafeDistance * self.options.playerSafeDistance) then
				--print("world spawn check failed: player = " .. playerX .. ":" .. playerY .. " - spawnpos = " ..spawnPos.x .. ":" .. spawnPos.y .. " distanceSquared: " .. distanceSquared .. " < " .. (self.options.playerSafeDistance * self.options.playerSafeDistance))
				satisfySafeSpawn = false
				break
			end
		end
		
		if satisfySafeSpawn then
			--print("world mob spawn pos: " .. spawnPos.x .. ":" .. spawnPos.y)
			return spawnPos.x, spawnPos.y
		else
			local attempt = (_attempt or 0) + 1
			if attempt < (self.options.maxSpawnAttempts or 10) then
				return world:getMobSpawnPosition(_movementType, attempt)
			else
				print("world: cannot spawn mob")
			end
			-- otherwise (too many attempts or just cannot find a spawn place) -> return nothing -> no spawn will happen
		end
	end
	
	
	function world:spawnMob(mobType)
		--print("world: spawn mob, type:")
		--print(mobType)
	
		local mob
		local offset = math.random(2, 12)
		local posX, posY
		
		-- note: now the spawn position is in world coordinates, not in grid coordinates!
		if mobType == "spider" then
			posX, posY = self:getMobSpawnPosition("walk")
			if posX == nil then return false end
			mob = spider:create(posX + offset, posY + offset)
			
		elseif mobType == "bat" then
			posX, posY = self:getMobSpawnPosition("walk")	-- TODO: movementType
			if posX == nil then return false end
			mob = bat:create(posX + offset, posY + offset)
			
		elseif mobType == "zombie" then
			posX, posY = self:getMobSpawnPosition("walk")
			if posX == nil then return false end
			mob = zombie:create(posX + offset, posY + offset)
			
		end
		
		-- note: the object gets its ID when the objmgr calls the objects load function (the objmgr creates the ID)
		--print("world: spawn mob, type:")
		--print(mobType)
		local id = objmgr:addObject(mob)
		self:addMob(id)
		self.lastMobSpawnTime = GAMETIME
		return true
		
	end
	
	
	function world:controlItemSpawn(dt)
	
	end
	
	
	function world:getClosestPlayerIDByPosition(_posX, _posY, _maxDistance)
		
		local distanceSquared, playerID = (_maxDistance * _maxDistance or 99999999), nil
		for _, _playerID in pairs(self.players) do
			local _player = objmgr:getObjectByID(_playerID)
			local pX, pY = _player:getPosition()
			local playerDistanceSquared = tools.getDistanceSquared({x = _posX, y = _posY}, {x = pX, y = pY})
			if playerDistanceSquared < distanceSquared then
				distanceSquared = playerDistanceSquared
				playerID = _player.id
			end
		end
		
		return playerID
		
	end
	
	
	function world:controlMobMovement(dt)
		for _, _mobID in pairs(self.mobs) do
			local mob = objmgr:getObjectByID(_mobID)
			if mob.behaviour == BEHAVIOUR_FOLLOW_PLAYER then
				local currentTarget = mob.currentTargetID
				
				local mobX, mobY = mob:getPosition()
				local closestPlayerID = self:getClosestPlayerIDByPosition(mobX, mobY, mob.followPlayerDistance)
				if closestPlayerID == nil then
					--print("world: no close player!")
					if currentTarget ~= nil then
						closestPlayerID = currentTarget
					end
				else
					if objmgr:getObjectByID(closestPlayerID) == nil then
						-- this player does not exist anymore / is dead -> forget the ID
						closestPlayerID = nil
					end
				end
				
				-- attack the attacker, or the closest target
				if closestPlayerID ~= nil then
					local closestPlayer = objmgr:getObjectByID(closestPlayerID)
					local goal = closestPlayer:getGridPosition()	-- table!
					local start = mob:getGridPosition()	-- table!
					if mob.pathToPlayer == nil then
						--print("start: " .. start.x .. ":" .. start.y)
						--print("goal: " .. goal.x .. ":" .. goal.y)
						
						-- already there? quit
						--if (start.x == goal.x) and (start.y == goal.y) then
							--return	--cannot use return here!
						--end
						
						local path = self.map:findPath(mob.movementType, start, goal)
						if path ~= nil then
							--print("world path 1 - " .. #path)
							table.remove(path, #path)	-- remove "first" entry because its the own position
							--print("world path 2 - " .. #path)
							mob.pathToPlayer = path
						end
					--else
					end
					
					if mob.pathToPlayer ~= nil then
						-- mob has a path to a player: TODO: there may be another player closer now...
						local pathLength = #mob.pathToPlayer
						--print("world: path length: " .. pathLength)
						
						-- TODO: if player moved to a grid pos that is already in the path; do not calculate path again!
						if pathLength < 1 then
							mob.pathToPlayer = nil
						else
							--print("world: path goal: " .. mob.pathToPlayer[1].x .. ":" .. mob.pathToPlayer[1].y)
							local gx, gy =  mob.pathToPlayer[1].x,  mob.pathToPlayer[1].y
							if (gx == goal.x) and (gy == goal.y) then
								-- target player is still there
								local next = mob.pathToPlayer[pathLength]
								--print("world set pos: [" .. next.x .. ":" .. next.y .. "] steps: " .. pathLength)
								-- TODO: actually change the movement direction, not the position
								--mob:setPosition(next.x * 64, next.y * 64)
								
								local dx, dy = (next.x - start.x) * mob.maxAcceleration, (next.y - start.y) * mob.maxAcceleration
								mob:accelerate(dt, dx, dy)
								table.remove(mob.pathToPlayer, pathLength)
							else
								-- target player has moved...
								mob.pathToPlayer = nil
							end
						end
					end
				end
				
			elseif mob.behaviour == BEHAVIOUR_WANDER_AROUND then
			
				-- randomly move around ... TODO: maybe use an angle instead
				local speedLimit = mob.maxAcceleration * 2
				local rndX, rndY = math.random(-speedLimit, speedLimit), math.random(-speedLimit, speedLimit)
				mob:accelerate(dt, rndX, rndY)
				--print("world: accelerate mob")
			
			elseif mob.behaviour == BEHAVIOUR_DO_CIRCLES then
			
			
			end
			
			mob:move(dt)
			
		end
	end
	
	-- TODO: remove the map objects
	function world:destroy()
		-- destroy everything inside this world
		for i = #self.mobs, 1, -1 do
			--self.mobs[i]:destroy()
			local mob = objmgr:getObjectByID(self.mobs[i])
			if mob then
				mob:destroy()
			end
		end
		
		for i = #self.players, 1, -1 do
			local ply = objmgr:getObjectByID(self.players[i])
			if ply then
				ply:destroy()
			end
		end
		
		self.map:destroy()
		
		gameloop:removeLoop(self)
		objmgr:destroyObject(self.id)
	end
	
	
	return world
	
end



return World